PRO BODY_LIMB, body, observer, et, npts, ra_limb, dec_limb, limberror
;; Given a NAIF body name, a NAIF observer name,
;; and a number of points, retrieve the RA and DEC of the 
;; body limb as seen from the observer at npts equally spaced
;; points around the limb.
;; Josh Colwell
;; July 12, 2004. Spot check with SOA successful.

ra_limb = FLTARR(npts)
dec_limb = FLTARR(npts)

limberror = 0B
CATCH, /CANCEL
CATCH, limb_error_status
IF STRMID(!ERROR_STATE.MSG, 0, 6) EQ 'CSPICE' THEN BEGIN
    limberror = 1B
    CATCH, /CANCEL
    MESSAGE, /RESET_ERROR_STATE
    RETURN
ENDIF

;; Get the state vector from the S/C to the target
CSPICE_SPKPOS, body, et, 'J2000', 'LT', observer, scpos, LT_CORR

CSPICE_BODN2C, body, bodyid, found
IF NOT found THEN PRINT,'ERROR: Body ', body,' not found in BODY_LIMB'
CSPICE_BODVAR, bodyid, 'RADII', radii       

;; Use the NAIF routine EDLIMB to find the limb of the body
;; from the pov of the S/C.

CSPICE_EDLIMB, RADII(0), RADII(1), RADII(2), SCPOS, BODLIMB
CSPICE_EL2CGV, BODLIMB, CENTER, V1, V2

;; Get RA/Dec points along the limb of the body:
FOR j = 0, npts-1 DO BEGIN
    AZ = !DPI*2.0/DOUBLE(npts)*j-!DPI
    VEC2 = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    limbvec = scpos-vec2
    CSPICE_RECRAD, LIMBVEC, RANGE, RA1, DEC1
    ra_limb[j] = RA1*!RADEG
    dec_limb[j] = DEC1 *!RADEG
ENDFOR

RETURN
END
