;; File Version: 1.2
;; Last Modified: December 18, 2007
;; By: AWD

PRO CALL_BODYMAP, mainuval, data, cube, geometer_base

COMPILE_OPT idl2

;; Procedure must be called by GEOMETER. Builds a temporary widget to 
;; collect user preferences for the DIRECT_BODYMAP plot. Put into this
;; separate procedure from GEOMETER_EVENT where it was originally 
;; developed.
;; Josh Colwell
;; July 31, 2006
;; January 16, 2007: Implemented graphic state vector.
;; June 7, 2007: Update for maptime internal to this widget, and 
;; geometer cube format.
;; December 18, 2007: Updated to allow multiple output file formats. 

bodymapbase = WIDGET_BASE(/COLUMN, TITLE = 'Direct Body Map Options')

IF mainuval.cubeloaded THEN limbtime = cube[0].et ELSE $
  limbtime = mainuval.etstart
targetid = WIDGET_INFO(mainuval.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, mainuval.targetdrop, GET_VALUE = $
                targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
BODY_LIMB, target, 'CASSINI', limbtime, 90, $
           ra_limb, dec_limb, limberror


;; Check to see if values from a previous run of target exist, and if so,
;; use those values.
IF mainuval.bodystate.use THEN BEGIN
   lonmin = mainuval.bodystate.lonmin
   lonmax = mainuval.bodystate.lonmax
   latmin = mainuval.bodystate.latmin
   latmax = mainuval.bodystate.latmax
ENDIF ELSE BEGIN
   lonmin = 0.
   lonmax = 360.
   latmin = -90.
   latmax = 90.
ENDELSE

lonbase = WIDGET_BASE(bodymapbase, /ROW)
lonminwidget = CW_FIELD(lonbase, TITLE = 'Min. longitude (deg): ',$
                       UVALUE = 'LONMINTEXT', VALUE = $
                       STRTRIM(STRING(lonmin),2), XSIZE = 12)
lonmaxwidget = CW_FIELD(lonbase, TITLE = 'Max. longitude (deg): ',$
                       UVALUE = 'LONMAXTEXT', VALUE = $
                       STRTRIM(STRING(lonmax),2), XSIZE = 12)
latbase = WIDGET_BASE(bodymapbase, /ROW)
latminwidget = CW_FIELD(latbase, TITLE = 'Min. latitude (deg): ',$
                       UVALUE = 'LATMINTEXT', VALUE= $
                       STRTRIM(STRING(latmin),2), XSIZE = 12)
latmaxwidget = CW_FIELD(latbase, TITLE = 'Max. latitude (deg): ',$
                       UVALUE = 'LATMAXTEXT', VALUE = $
                       STRTRIM(STRING(latmax),2), XSIZE = 12)

;; Check to see if values from a previous run of target exist, and if so,
;; use those values.
IF mainuval.bodystate.use THEN BEGIN
   WIDGET_CONTROL, lonminwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.bodystate.lonmin),2)
   WIDGET_CONTROL, lonmaxwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.bodystate.lonmax),2)
   WIDGET_CONTROL, latminwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.bodystate.latmin),2)
   WIDGET_CONTROL, latmaxwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.bodystate.latmax),2)
ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
   specminred = '0'
   specmaxred = '0'
   specmingreen = '0'
   specmaxgreen = '0'
   specminblue = '0'
   specmaxblue = '0'
ENDIF ELSE BEGIN
   nspec = N_ELEMENTS(WHERE(cube[0].data[0,*] NE -1.))
   IF mainuval.bodystate.use THEN BEGIN
      specminred = STRTRIM(STRING(mainuval.bodystate.spec_bounds[0]),2)
      specmaxred = STRTRIM(STRING(mainuval.bodystate.spec_bounds[1]),2)
      specmingreen = STRTRIM(STRING(mainuval.bodystate.spec_bounds[2]),2)
      specmaxgreen = STRTRIM(STRING(mainuval.bodystate.spec_bounds[3]),2)
      specminblue = STRTRIM(STRING(mainuval.bodystate.spec_bounds[4]),2)
      specmaxblue = STRTRIM(STRING(mainuval.bodystate.spec_bounds[5]),2)
   ENDIF ELSE BEGIN
      specminred = '0'
      specmaxred = STRTRIM(STRING(nspec/3),2)
      specmingreen = STRTRIM(STRING(nspec/3+1),2)
      specmaxgreen = STRTRIM(STRING(nspec-nspec/3),2)
      specminblue = STRTRIM(STRING(nspec-nspec/3+1),2)
      specmaxblue = STRTRIM(STRING(nspec-1),2)
   ENDELSE
ENDELSE

plotspecbase = WIDGET_BASE(bodymapbase, /ROW, /FRAME)
plotrowselect = CW_FIELD(plotspecbase, TITLE = 'Plot Row: ', $
                         UVALUE = 'PLOTROW', VALUE = '0', XSIZE = 4)
plottimeselect = CW_FIELD(plotspecbase, TITLE = 'Plot Index: ', $
                          UVALUE = 'PLOTTIME', VALUE = '0', XSIZE =4)
plotbutton = WIDGET_BUTTON(plotspecbase, VALUE = 'Plot Spectrum', $
                           UVALUE = 'PLOTBUTTON')
IF NOT mainuval.cubeloaded THEN WIDGET_CONTROL, plotspecbase, SENSITIVE = 0

redbase = WIDGET_BASE(bodymapbase, /ROW)
redwidget = CW_FIELD(redbase, TITLE = 'Red channels (lo,hi): ',$
                     UVALUE = 'REDTEXT', VALUE = specminred+','+specmaxred, $
                     XSIZE = 8)
redmaxwidget = CW_FIELD(redbase, TITLE = 'Red maximum: ', UVALUE = $
                        'REDMAX', VALUE = '0', XSIZE = 6)

greenbase = WIDGET_BASE(bodymapbase, /ROW)
greenwidget = CW_FIELD(greenbase, TITLE = 'Green channels (lo,hi): ',$
                       UVALUE = 'GREENTEXT', VALUE = specmingreen+','+$
                       specmaxgreen, XSIZE = 8)
greenmaxwidget = CW_FIELD(greenbase, TITLE = 'Green maximum: ', UVALUE = $
                        'GREENMAX', VALUE = '0', XSIZE = 6)

bluebase = WIDGET_BASE(bodymapbase, /ROW)
bluewidget = CW_FIELD(bluebase, TITLE = 'Blue channels (lo,hi): ',$
                      UVALUE = 'BLUETEXT', VALUE = specminblue+','+$
                      specmaxblue, XSIZE = 8)
bluemaxwidget = CW_FIELD(bluebase, TITLE = 'Blue maximum: ', UVALUE = $
                        'BLUEMAX', VALUE = '0', XSIZE = 6)

;; Allow user to choose map projection
baseforframe2 = WIDGET_BASE(bodymapbase, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe2, VALUE = 'Map projection:')
gridbase = WIDGET_BASE(baseforframe2, /ROW, /EXCLUSIVE)
gridbutton = WIDGET_BUTTON(gridbase, VALUE = $
                               'None', $
                               UVALUE = 'GRID')
aznbutton = WIDGET_BUTTON(gridbase, VALUE = $
                              'Azimuthal (N)', $
                              UVALUE = 'AZN')
azsbutton = WIDGET_BUTTON(gridbase, VALUE = $
                              'Azimuthal (S)', $
                              UVALUE = 'AZS')
stebutton = WIDGET_BUTTON(gridbase, VALUE = $
                              'Stereo (E)', $
                              UVALUE = 'STE')
stwbutton = WIDGET_BUTTON(gridbase, VALUE = $
                              'Stereo (W)', $
                              UVALUE = 'STW')
WIDGET_CONTROL, gridbutton, /SET_BUTTON
grid = 1B
azn = 0B
azs = 0B
ste = 0B
stw = 0B

;; Allow user to select subset of available times:
timeselectbase = WIDGET_BASE(bodymapbase, /ROW, /FRAME)
ntimewidget = CW_FIELD(timeselectbase, TITLE = 'Number of time steps to plot', $
                       VALUE = '1', UVALUE = 'TIMETEXT', XSIZE = 8)

;; Have a series of Yes/No buttons for various Sky Map options.
;; Allow user to show all FOV pointings at once (but a single
;; background).
timesbase = WIDGET_BASE(bodymapbase, /ROW, /EXCLUSIVE)
alltimesbutton = WIDGET_BUTTON(timesbase, VALUE = $
                               'Show all times', $
                               UVALUE = 'ALLTIMES')
onetimebutton = WIDGET_BUTTON(timesbase, VALUE = $
                              'Show selected time', $
                              UVALUE = 'ONETIME')
WIDGET_CONTROL, onetimebutton, /SET_BUTTON
alltimes = 0B

;; Allow user to select an animation:
animbase = WIDGET_BASE(bodymapbase, /ROW, /EXCLUSIVE)
animyes = WIDGET_BUTTON(animbase, VALUE = 'Animate', $
                        UVALUE = 'ANIMYES')
animno = WIDGET_BUTTON(animbase, VALUE = 'Single Frame', $
                       UVALUE = 'ANIMNO')
WIDGET_CONTROL, animno, /SET_BUTTON
bodyanim = 0B

;; Allow user to choose a specific time at which to draw map
mapbase = WIDGET_BASE(bodymapbase, /ROW, /FRAME)
IF mainuval.cubeloaded THEN $
  CSPICE_ET2UTC, cube[0].et, 'ISOD', 3L, startutc ELSE $
    CSPICE_ET2UTC, mainuval.etstart, 'ISOD', 3L, startutc
maptimewidget = CW_FIELD(mapbase, TITLE = 'Map time', $
                         VALUE = startutc, $
                         UVALUE = 'MAPTIMETEXT', XSIZE = 21, $
                        /RETURN_EVENTS)

;; Allow user to use data cube times (corresponding to 
;; data records) rather than entered delta-times:
cubebase = WIDGET_BASE(bodymapbase, /ROW, /EXCLUSIVE)
cubeyes = WIDGET_BUTTON(cubebase, VALUE = 'Use Data Times', $
                        UVALUE = 'CUBEYES')
cubeno = WIDGET_BUTTON(cubebase, VALUE = 'Use Entered Times', $
                       UVALUE = 'CUBENO', SENSITIVE=0)
IF(N_ELEMENTS(data) NE 0) THEN $
                    WIDGET_CONTROL, cubeno, SENSITIVE = 1
cubetimes = 0B

;; Allow user to select screen, PS, JPG, or TIFF output and filename:
outputnames = ['Screen', 'PS', 'JPG', 'TIFF']
outputbuttons = CW_BGROUP(bodymapbase, outputnames, UVALUE = 'OUTPUTBUTTON', $
                    /ROW, /EXCLUSIVE, /RETURN_NAME, SET_VALUE = 0, $
                    UNAME = 'outputbutton', LABEL_LEFT = 'Output type')
savefilewidget = CW_FIELD(bodymapbase, TITLE = 'Filename:', UVALUE = $
                        'SAVEFILE', VALUE = 'bodymap', XSIZE = 20)
screen = 1B
postscript = 0B
jpg = 0B
tiff = 0B

IF mainuval.bodystate.use THEN BEGIN
   savefile = STRTRIM(mainuval.bodystate.savefile,2)
   WIDGET_CONTROL, savefilewidget, SET_VALUE = savefile
ENDIF ELSE savefile = ''

IF NOT MAINUVAL.bodystate.use THEN BEGIN
    mainuval.bodystate.grid = grid
    mainuval.bodystate.azn = azn
    mainuval.bodystate.azs = azs
    mainuval.bodystate.ste = ste
    mainuval.bodystate.stw = stw
    mainuval.bodystate.alltimes = alltimes
    mainuval.bodystate.animate = bodyanim
    mainuval.bodystate.cubetimes = cubetimes
    mainuval.bodystate.screen = screen
    mainuval.bodystate.postscript = postscript
    mainuval.bodystate.jpg = jpg
    mainuval.bodystate.tiff = tiff
    mainuval.bodystate.savefile = savefile
ENDIF

IF mainuval.bodystate.use THEN BEGIN
    WIDGET_CONTROL, redmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.bodystate.spec_max[0]),2)
    WIDGET_CONTROL, greenmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.bodystate.spec_max[1]),2)
    WIDGET_CONTROL, bluemaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.bodystate.spec_max[2]),2)
    WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.bodystate.ntime),2)
    WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.bodystate.maptime),2)

    IF mainuval.bodystate.grid THEN BEGIN
       WIDGET_CONTROL, gridbutton, /SET_BUTTON
       grid = 1B
       azn = 0B
       azs = 0B
       ste = 0B
       stw = 0B
    ENDIF
    IF mainuval.bodystate.azn THEN BEGIN
       WIDGET_CONTROL, aznbutton, /SET_BUTTON
       azn = 1B
       grid = 0B
       azs = 0B
       ste = 0B
       stw = 0B
    ENDIF
    IF mainuval.bodystate.azs THEN BEGIN
       WIDGET_CONTROL, azsbutton, /SET_BUTTON
       azs = 1B
       grid = 0B
       azn = 0B
       ste = 0B
       stw = 0B
    ENDIF
    IF mainuval.bodystate.ste THEN BEGIN
       WIDGET_CONTROL, stebutton, /SET_BUTTON
       ste = 1B
       grid = 0B
       azn = 0B
       azs = 0B
       stw = 0B
    ENDIF
    IF mainuval.bodystate.stw THEN BEGIN
       WIDGET_CONTROL, stwbutton, /SET_BUTTON
       stw = 1B
       grid = 0B
       azn = 0B
       azs = 0B
       ste = 0B
    ENDIF

    IF mainuval.bodystate.animate THEN BEGIN
       WIDGET_CONTROL, animyes, /SET_BUTTON
       bodyanim = 1B
    ENDIF
    IF mainuval.bodystate.alltimes THEN BEGIN
       WIDGET_CONTROL, alltimesbutton, /SET_BUTTON
       alltimes = 1B
    ENDIF
    IF mainuval.bodystate.cubetimes THEN BEGIN
       WIDGET_CONTROL, cubeyes, /SET_BUTTON
       cubetimes = 1B
    ENDIF

    IF mainuval.bodystate.screen THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
       screen = 1B
    ENDIF
    IF mainuval.bodystate.postscript THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 1
       postscript = 1B
       savefile = mainuval.bodystate.savefile
    ENDIF
    IF mainuval.bodystate.jpg THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 2
       jpg = 1B
       savefile = mainuval.bodystate.savefile
    ENDIF
    IF mainuval.bodystate.tiff THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 3
       tiff = 1B
       savefile = mainuval.bodystate.savefile
    ENDIF

ENDIF

buttonbase = WIDGET_BASE(bodymapbase, /ROW)
helpbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Help', $
                           UVALUE = 'HELPBUTTON')
cancelbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Cancel', $
                             UVALUE = 'CANCELBUTTON')

;; Clicking the done button on this widget results in all 
;; values being read from the widget and passed to
;; DIRECT_BODYMAP.
donebutton = WIDGET_BUTTON(buttonbase, VALUE = 'Make it so', $
                                 UVALUE = 'DONEBUTTON', SENSITIVE=0)

WIDGET_CONTROL, bodymapbase, SET_UVALUE = bodymapuval
WIDGET_CONTROL, bodymapbase, /REALIZE, GROUP = geometer_base, /MODAL
IF mainuval.bodystate.cubetimes THEN $
       WIDGET_CONTROL, donebutton, SENSITIVE = 1

;; Check this widget to see if the Done button is pressed.
;; Once the done button is pressed, read the values from the
;; various inputs in this widget, destroy the widget, and then
;; send these inputs to the bodymap routine.

done_with_bodymapinput = 0B
spec_bounds = [0,1,2,3,4,5]

REPEAT BEGIN
   bodymapevent = WIDGET_EVENT()
   CASE bodymapevent.top OF
      bodymapbase:BEGIN
         WIDGET_CONTROL, bodymapevent.id, GET_UVALUE = ev_uval
         CASE ev_uval OF
            'HELPBUTTON': XDISPLAYFILE, mainuval.geometerpath+$
                'directbodymaphelp.txt', DONE_BUTTON = 'Done', $
                                        GROUP = bodymapbase, /MODAL
            'CANCELBUTTON': BEGIN
               WIDGET_CONTROL, bodymapevent.top, /DESTROY
               RETURN
            END
            'DONEBUTTON':BEGIN
               WIDGET_CONTROL, lonminwidget, GET_VALUE = lonmin
               lonmin = FLOAT(lonmin[0])
               mainuval.bodystate.lonmin = lonmin
               WIDGET_CONTROL, lonmaxwidget, GET_VALUE = lonmax
               lonmax = FLOAT(lonmax[0])
               mainuval.bodystate.lonmax = lonmax
               WIDGET_CONTROL, latminwidget, GET_VALUE = latmin
               latmin = FLOAT(latmin[0])
               mainuval.bodystate.latmin = latmin
               WIDGET_CONTROL, latmaxwidget, GET_VALUE = latmax
               latmax = FLOAT(latmax[0])
               mainuval.bodystate.latmax = latmax

               ;; Read out the color selections:
               WIDGET_CONTROL, redwidget, GET_VALUE = red
               redlims = STRSPLIT(red,',',/EXTRACT)
               spec_bounds[0] = FIX(redlims[0])
               spec_bounds[1] = FIX(redlims[1])
               WIDGET_CONTROL, redmaxwidget, GET_VALUE = redmaxstring
               redmax = FLOAT(redmaxstring[0])
               
               WIDGET_CONTROL, greenwidget, GET_VALUE = green
               greenlims = STRSPLIT(green,',',/EXTRACT)
               spec_bounds[2] = FIX(greenlims[0])
               spec_bounds[3] = FIX(greenlims[1])
               WIDGET_CONTROL, greenmaxwidget, GET_VALUE = greenmaxstring
               greenmax = FLOAT(greenmaxstring[0])

               WIDGET_CONTROL, bluewidget, GET_VALUE = blue
               bluelims = STRSPLIT(blue,',',/EXTRACT)
               spec_bounds[4] = FIX(bluelims[0])
               spec_bounds[5] = FIX(bluelims[1])
               WIDGET_CONTROL, bluemaxwidget, GET_VALUE = bluemaxstring
               bluemax = FLOAT(bluemaxstring[0])

               mainuval.bodystate.spec_bounds = spec_bounds
               mainuval.bodystate.spec_max = [redmax, greenmax, bluemax]

               WIDGET_CONTROL, ntimewidget, GET_VALUE = ntimestring
               ntimeplot = FIX(FLOAT(ntimestring[0]))
               mainuval.bodystate.ntime = ntimeplot

               WIDGET_CONTROL, maptimewidget, GET_VALUE = maptimestring
               maptime = STRTRIM(maptimestring[0],2)
               mainuval.bodystate.maptime = maptime

               WIDGET_CONTROL, savefilewidget, GET_VALUE = savefile
               mainuval.bodystate.savefile = savefile

               done_with_bodymapinput = 1B
               mainuval.bodystate.use = 1B
            END
            'PLOTBUTTON': BEGIN
               WIDGET_CONTROL, plotrowselect, GET_VALUE = rowselect
               rowselect = FIX(FLOAT(rowselect[0]))
               WIDGET_CONTROL, plottimeselect, GET_VALUE = timeselect
               timeselect = FIX(FLOAT(timeselect[0]))
               PLOT, cube[timeselect].data[rowselect,*], XTITLE = $
                     'Channel Number', YTITLE = 'Data Cube Value'
            END
            'ALLTIMES': BEGIN
               alltimes = 1B
               mainuval.bodystate.alltimes = 1B
            END
            'ONETIME': BEGIN
               alltimes = 0B
               mainuval.bodystate.alltimes = 0B
            END
            'OUTPUTBUTTON': BEGIN
               WIDGET_CONTROL, outputbuttons, GET_VALUE = outputtype
               CASE outputtype OF
                  0: BEGIN
                     screen = 1B
                     postscript = 0B
                     jpg = 0B
                     tiff = 0B
                     mainuval.bodystate.screen = 1B
                     mainuval.bodystate.postscript = 0B
                     mainuval.bodystate.jpg = 0B
                     mainuval.bodystate.tiff = 0B
                  END
                  1: BEGIN
                     screen = 0B
                     postscript = 1B
                     jpg = 0B
                     tiff = 0B
                     mainuval.bodystate.screen = 0B
                     mainuval.bodystate.postscript = 1B
                     mainuval.bodystate.jpg = 0B
                     mainuval.bodystate.tiff = 0B
                     mainuval.bodystate.savefile = savefile
                  END
                  2: BEGIN
                     screen = 0B
                     jpg = 1B
                     postscript = 0B
                     tiff = 0B
                     mainuval.bodystate.screen = 0B
                     mainuval.bodystate.postscript = 0B
                     mainuval.bodystate.jpg = 1B
                     mainuval.bodystate.tiff = 0B
                     mainuval.bodystate.savefile = savefile
                  END
                  3: BEGIN
                     screen = 0B
                     jpg = 0B
                     postscript = 0B
                     tiff = 1B
                     mainuval.bodystate.screen = 0B
                     mainuval.bodystate.postscript = 0B
                     mainuval.bodystate.jpg = 0B
                     mainuval.bodystate.tiff = 1B
                     mainuval.bodystate.savefile = savefile
                  END
               ENDCASE
            END
            'ANIMYES': BEGIN
               bodyanim = 1B
               mainuval.bodystate.animate = 1B
               screen = 1B
               postscript = 0B
               jpg = 0B
               tiff = 0B
               WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
            END
            'ANIMNO': BEGIN
               bodyanim = 0B
               mainuval.bodystate.animate = 0B
            END
            'GRID': BEGIN
               grid = 1B
               mainuval.bodystate.grid = 1B
               mainuval.bodystate.azn = 0B
               mainuval.bodystate.azs = 0B
               mainuval.bodystate.ste = 0B
               mainuval.bodystate.stw = 0B
            END
            'AZN': BEGIN
               azn = 1B
               grid = 0B
               mainuval.bodystate.azn = 1B
               mainuval.bodystate.grid = 0B
               mainuval.bodystate.azs = 0B
               mainuval.bodystate.ste = 0B
               mainuval.bodystate.stw = 0B
            END
            'AZS': BEGIN
               azs = 1B
               grid = 0B
               mainuval.bodystate.azs = 1B
               mainuval.bodystate.grid = 0B
               mainuval.bodystate.azn = 0B
               mainuval.bodystate.ste = 0B
               mainuval.bodystate.stw = 0B
            END
            'STE': BEGIN
               ste = 1B
               grid = 0B
               mainuval.bodystate.ste = 1B
               mainuval.bodystate.grid = 0B
               mainuval.bodystate.azn = 0B
               mainuval.bodystate.azs = 0B
               mainuval.bodystate.stw = 0B
            END
            'STW': BEGIN
               stw = 1B
               grid = 0B
               mainuval.bodystate.stw = 1B
               mainuval.bodystate.grid = 0B
               mainuval.bodystate.azn = 0B
               mainuval.bodystate.azs = 0B
               mainuval.bodystate.ste = 0B
            END
            'MAPTIMETEXT': BEGIN
               WIDGET_CONTROL, maptimewidget, GET_VALUE = $
                               maptimestring
               CSPICE_UTC2ET, maptimestring[0], maptimeet
               CSPICE_ET2UTC, maptimeet, 'ISOD', 3L, maptimestring
               WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                               maptimestring
            END
            'CUBEYES': BEGIN
               cubetimes = 1B
               mainuval.bodystate.cubetimes = 1B
               limbtime = cube.et[0,0]
               ntimes = N_ELEMENTS(cube.et)
               IF NOT mainuval.bodystate.use THEN $
                  WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                                  STRING(STRTRIM(ntimes),2)
               BODY_LIMB, target, 'CASSINI', limbtime, 90, $
                          ra_limb, dec_limb, limberror
               ;; Allow user to continue
               WIDGET_CONTROL, donebutton, SENSITIVE = 1

            END
            'CUBENO': BEGIN
               IF N_ELEMENTS(data) NE 0 THEN BEGIN
                  limbtime = data[0].et
                  ntimes = N_ELEMENTS(data.ra_full[0])
                  WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                                  STRING(STRTRIM(ntimes),2)
                  BODY_LIMB, target, 'CASSINI', limbtime, 90, $
                             ra_limb, dec_limb, limberror
                  ;; Allow user to continue
                  WIDGET_CONTROL, donebutton, SENSITIVE = 1

                  cubetimes = 0B
                  mainuval.bodystate.cubetimes = 0B
               ENDIF            ; check that data is defined
            END
            ELSE:
         ENDCASE
      END
      ELSE: ;; Do nothing
   ENDCASE
ENDREP UNTIL done_with_bodymapinput
WIDGET_CONTROL, bodymapbase, /DESTROY

IF cubetimes THEN ntimes = N_ELEMENTS(cube.et) ELSE $
   ntimes = N_ELEMENTS(data.ra_full[0])

IF bodyanim THEN XINTERANIMATE, SET = [600, 600, ntimeplot] 

;; Identify the time to make the map from the user entry:
IF NOT cubetimes THEN BEGIN
;    WIDGET_CONTROL, mainuval.maptimetext, GET_VALUE = maptime
;    maptime = maptime[0]
    CSPICE_UTC2ET, maptime, mapet
    junk = MIN(ABS(mapet-data.et),etloc)
    CSPICE_ET2UTC, data[etloc].et, mainuval.tformat, 3L, maptime
    mainuval.bodystate.maptime = maptime
    WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                    'Actual time for map: ' + maptime
ENDIF ELSE BEGIN
   CSPICE_UTC2ET, maptime, mapet
   junk = MIN(ABS(mapet-cube.et), etloc)
   CSPICE_ET2UTC, cube[etloc].et, mainuval.tformat, 3L, maptime
   mainuval.bodystate.maptime = maptime
ENDELSE

;; Determine which set of times to use and store in vectors
;; to pass to DIRECT_BODYMAP.

dtime_ix = FLOAT(ntimes)/(ntimeplot-1)
plot_ets = DBLARR(ntimeplot)
plot_utcs = STRARR(ntimeplot)

IF cubetimes THEN BEGIN
    IF mainuval.cubeloaded THEN BEGIN
        cube_bodymap = cube
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = cube[gettime_ix].et
            CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                           3L, utc0
            plot_utcs[time_ix] = utc0
        ENDFOR
    ENDIF ELSE BEGIN
        WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                        'Cube not loaded.'
        WIDGET_CONTROL, bodymapbase, /DESTROY
        RETURN
    ENDELSE
ENDIF ELSE BEGIN
    ;; Create a dummy variable to take the place of the cube
    ;; data to pass to DIRECT_BODYMAP since cube data may
    ;; not be loaded if we're here.
    cube_bodymap = {et:FLTARR(1)+1E20} 
    ;; Give it a ridiculous value so
    ;; DIRECT_BODYMAP will know it's not real.

    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        gettime_ix = time_ix*dtime_ix < (ntimes-1)
        plot_ets[time_ix] = data[gettime_ix].et
        CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                       3L, utc0
        plot_utcs[time_ix] = utc0
    ENDFOR
ENDELSE

;; Now make the call(s) to DIRECT_BODYMAP
IF bodyanim THEN BEGIN
   maptimedummy = ''
    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        data_ix = time_ix*dtime_ix < (ntimes-1)
        DIRECT_BODYMAP, data, mainuval, data_ix, time_ix, plot_ets, $
                        plot_utcs, cube_bodymap, spec_bounds, $
                        lonmin, lonmax, latmin, latmax, maptimedummy, $
                        COLORMAX = [redmax, greenmax, bluemax], $
                        ALLTIMES = alltimes, /ANIMATE, GRID = grid, $
                        AZN = azn, AZS = azs, STE = ste, STW = stw, $
                        SCREEN = screen, POSTSCRIPT = postscript, $
                        JPG = jpg, TIFF = tiff, SAVEFILE = savefile
        XINTERANIMATE, FRAME = time_ix, WINDOW = !D.WINDOW
    ENDFOR
    XINTERANIMATE
ENDIF ELSE $
  DIRECT_BODYMAP, data, mainuval, etloc, etloc, plot_ets, $
                  plot_utcs, cube_bodymap, spec_bounds, $
                  lonmin, lonmax, latmin, latmax, maptime, $
                  COLORMAX = [redmax, greenmax, bluemax], $
                  ALLTIMES = alltimes, GRID = grid, $
                  AZN = azn, AZS = azs, STE = ste, STW = stw, $
                  SCREEN = screen, POSTSCRIPT = postscript, $
                  JPG = jpg, TIFF = tiff, SAVEFILE = savefile

END
