;; File Version: 1.4
;; Last Modified: June 2, 2009
;; By: JEC

PRO CALL_SKYMAP, mainuval, data, cube, geometer_base

COMPILE_OPT idl2

;; Procedure must be called by GEOMETER. Builds a temporary widget to 
;; collect user preferences for the DIRECT_SKYMAP plot. Put into this
;; separate procedure from GEOMETER_EVENT where it was originally 
;; developed.
;; Josh Colwell
;; July 31, 2006
;;
;; January 16, 2007: Added graphic state vector. AWD.
;; April 3, 2007: Modify for GEOMETER version of data cube. Get map time
;; working.
;; June 6, 2007: Sync up with direct_target, and internal (to this
;; widget) map time handling. Add data info button. Handle new cube format.
;; December 18, 2007: Updated to allow multiple output file formats. 
;; June 2, 2009: Fix animation bug.

;; Create a window to get plotting preferences.
skymapbase = WIDGET_BASE(/COLUMN, TITLE = 'Direct Sky Map Options')

;; Check zoom for default RA/Dec values:
;; Default is full sky. 
ra_min = 360.
ra_max = 0.
dec_min = -90.
dec_max = 90.
;; Get the target body limb limits and double it for the 
;; "zoomed in" option (triple if target is Saturn to 
;; allow room for rings).

IF mainuval.cubeloaded THEN limbtime = cube[0].et ELSE $
  limbtime = mainuval.etstart
targetid = WIDGET_INFO(mainuval.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, mainuval.targetdrop, GET_VALUE = $
                targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
BODY_LIMB, target, 'CASSINI', limbtime, 90, $
           ra_limb, dec_limb, limberror

IF mainuval.skyzoom EQ 2 THEN BEGIN
    IF target EQ 'SATURN' THEN zoomscalefactor = 3. ELSE $
      zoomscalefactor = 2.
    ra_min = MAX(ra_limb)
    ra_max = MIN(ra_limb)
    ra_center = ra_min-.5*ABS(ra_max-ra_min)
    delta_ra = zoomscalefactor*ABS(ra_max-ra_min) < 360.
    dec_min = MIN(dec_limb)
    dec_max = MAX(dec_limb)
    dec_center = dec_min+.5*(dec_max-dec_min)
    delta_dec = zoomscalefactor*(dec_max-dec_min) < 180.
    
    ra_min = (ra_center+.5*delta_ra) > 0.
    ra_max = (ra_min-delta_ra) < 360.
    dec_min = (dec_center-.5*delta_dec) > (-90.)
    dec_max = (dec_min+delta_dec) < 90.
ENDIF

raminwidget = CW_FIELD(skymapbase, TITLE = 'Minimum RA (deg): ',$
                       UVALUE = 'RAMINTEXT', VALUE = $
                       STRTRIM(STRING(ra_min),2), XSIZE = 12)
ramaxwidget = CW_FIELD(skymapbase, TITLE = 'Maximum RA (deg): ',$
                       UVALUE = 'RAMAXTEXT', VALUE = $
                       STRTRIM(STRING(ra_max),2), XSIZE = 12)
deminwidget = CW_FIELD(skymapbase, TITLE = 'Minimum Dec (deg): ',$
                       UVALUE = 'DEMINTEXT', VALUE= $
                       STRTRIM(STRING(dec_min),2), XSIZE = 12)
demaxwidget = CW_FIELD(skymapbase, TITLE = 'Maximum Dec (deg): ',$
                       UVALUE = 'DEMAXTEXT', VALUE = $
                       STRTRIM(STRING(dec_max),2), XSIZE = 12)

;; Check to see if values from a previous run of target exist, and if so,
;; use those values.
IF mainuval.skystate.use THEN BEGIN
   WIDGET_CONTROL, raminwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.skystate.ramin),2)
   WIDGET_CONTROL, ramaxwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.skystate.ramax),2)
   WIDGET_CONTROL, deminwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.skystate.decmin),2)
   WIDGET_CONTROL, demaxwidget, SET_VALUE = $
                   STRTRIM(STRING(mainuval.skystate.decmax),2)
ENDIF


IF cube[0].et GT 1E19 THEN BEGIN
    specminred = '0'
    specmaxred = '0'
    specmingreen = '0'
    specmaxgreen = '0'
    specminblue = '0'
    specmaxblue = '0'
ENDIF ELSE BEGIN
   nspec = N_ELEMENTS(WHERE(cube[0].data[0,*] NE -1.))
   IF mainuval.skystate.use THEN BEGIN
      specminred = STRTRIM(STRING(mainuval.skystate.spec_bounds[0]),2)
      specmaxred = STRTRIM(STRING(mainuval.skystate.spec_bounds[1]),2)
      specmingreen = STRTRIM(STRING(mainuval.skystate.spec_bounds[2]),2)
      specmaxgreen = STRTRIM(STRING(mainuval.skystate.spec_bounds[3]),2)
      specminblue = STRTRIM(STRING(mainuval.skystate.spec_bounds[4]),2)
      specmaxblue = STRTRIM(STRING(mainuval.skystate.spec_bounds[5]),2)
   ENDIF ELSE BEGIN
      specminred = '0'
      specmaxred = STRTRIM(STRING(nspec/3),2)
      specmingreen = STRTRIM(STRING(nspec/3+1),2)
      specmaxgreen = STRTRIM(STRING(nspec-nspec/3),2)
      specminblue = STRTRIM(STRING(nspec-nspec/3+1),2)
      specmaxblue = STRTRIM(STRING(nspec-1),2)
   ENDELSE
ENDELSE

plotspecbase = WIDGET_BASE(skymapbase, /ROW, /FRAME)
plotrowselect = CW_FIELD(plotspecbase, TITLE = 'Plot Row: ', $
                         UVALUE = 'PLOTROW', VALUE = '0', XSIZE = 4)
plottimeselect = CW_FIELD(plotspecbase, TITLE = 'Plot Index: ', $
                          UVALUE = 'PLOTTIME', VALUE = '0', XSIZE =4)
plotbutton = WIDGET_BUTTON(plotspecbase, VALUE = 'Plot Spectrum', $
                           UVALUE = 'PLOTBUTTON')
infobutton = WIDGET_BUTTON(plotspecbase, VALUE = 'Data Info', UVALUE = $
                           'DATAINFO')
IF NOT mainuval.cubeloaded THEN WIDGET_CONTROL, plotspecbase, SENSITIVE = 0

redbase = WIDGET_BASE(skymapbase, /ROW)
redwidget = CW_FIELD(redbase, TITLE = 'Red channels (lo,hi): ',$
                     UVALUE = 'REDTEXT', VALUE = specminred+','+specmaxred, $
                     XSIZE = 8)
redmaxwidget = CW_FIELD(redbase, TITLE = 'Red maximum: ', UVALUE = $
                        'REDMAX', VALUE = '0', XSIZE = 6)

greenbase = WIDGET_BASE(skymapbase, /ROW)
greenwidget = CW_FIELD(greenbase, TITLE = 'Green channels (lo,hi): ',$
                       UVALUE = 'GREENTEXT', VALUE = specmingreen+','+$
                       specmaxgreen, XSIZE = 8)
greenmaxwidget = CW_FIELD(greenbase, TITLE = 'Green maximum: ', UVALUE = $
                        'GREENMAX', VALUE = '0', XSIZE = 6)

bluebase = WIDGET_BASE(skymapbase, /ROW)
bluewidget = CW_FIELD(bluebase, TITLE = 'Blue channels (lo,hi): ',$
                      UVALUE = 'BLUETEXT', VALUE = specminblue+','+$
                      specmaxblue, XSIZE = 8)
bluemaxwidget = CW_FIELD(bluebase, TITLE = 'Blue maximum: ', UVALUE = $
                        'BLUEMAX', VALUE = '0', XSIZE = 6)

;; Allow user to choose whether to plot lat/lon lines
baseforframe2 = WIDGET_BASE(skymapbase, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe2, VALUE = 'Show Lat/Lon of target: ')
latlonbase = WIDGET_BASE(baseforframe2, /ROW, /EXCLUSIVE)
showlatlonbutton = WIDGET_BUTTON(latlonbase, VALUE = 'Yes', $
                                 UVALUE = 'LATLONYES')
hidelatlonbutton = WIDGET_BUTTON(latlonbase, VALUE = 'No', $
                                 UVALUE = 'LATLONNO')
WIDGET_CONTROL, showlatlonbutton, /SET_BUTTON
latlon = 1B

;; Allow user to select subset of available times:
timeselectbase = WIDGET_BASE(skymapbase, /ROW, /FRAME)
ntimewidget = CW_FIELD(timeselectbase, TITLE = 'Number of time steps to plot', $
                       VALUE = '1', UVALUE = 'TIMETEXT', XSIZE = 8)

;; Have a series of Yes/No buttons for various Sky Map options.
;; Allow user to show all FOV pointings at once (but a single
;; background).
timesbase = WIDGET_BASE(skymapbase, /ROW, /EXCLUSIVE)
alltimesbutton = WIDGET_BUTTON(timesbase, VALUE = $
                               'Show all times', $
                               UVALUE = 'ALLTIMES')
onetimebutton = WIDGET_BUTTON(timesbase, VALUE = $
                              'Show selected time', $
                              UVALUE = 'ONETIME')
WIDGET_CONTROL, onetimebutton, /SET_BUTTON
alltimes = 0B

;; Allow user to select an animation:
animbase = WIDGET_BASE(skymapbase, /ROW, /EXCLUSIVE)
animyes = WIDGET_BUTTON(animbase, VALUE = 'Animate', $
                        UVALUE = 'ANIMYES')
animno = WIDGET_BUTTON(animbase, VALUE = 'Single Frame', $
                       UVALUE = 'ANIMNO')
WIDGET_CONTROL, animno, /SET_BUTTON
skyanim = 0B

;; Allow user to use data cube times (corresponding to 
;; data records) rather than entered delta-times:
cubebase = WIDGET_BASE(skymapbase, /ROW, /EXCLUSIVE)
cubeyes = WIDGET_BUTTON(cubebase, VALUE = 'Use Data Times', $
                        UVALUE = 'CUBEYES')
cubeno = WIDGET_BUTTON(cubebase, VALUE = 'Use Entered Times', $
                       UVALUE = 'CUBENO', SENSITIVE=0)
IF(N_ELEMENTS(data) NE 0) THEN $
                    WIDGET_CONTROL, cubeno, SENSITIVE = 1
cubetimes = 0B

;; Allow user to choose a specific time at which to draw map
mapbase = WIDGET_BASE(skymapbase, /ROW, /FRAME)
IF mainuval.cubeloaded THEN $
  CSPICE_ET2UTC, cube[0].et, 'ISOD', 3L, startutc ELSE $
    CSPICE_ET2UTC, mainuval.etstart, 'ISOD', 3L, startutc
maptimewidget = CW_FIELD(mapbase, TITLE = 'Map time', $
                         VALUE = startutc, $
                         UVALUE = 'MAPTIMETEXT', XSIZE = 21, $
                        /RETURN_EVENTS)

;; Allow user to select screen, PS, JPG, or TIFF output and filename:
outputnames = ['Screen', 'PS', 'JPG', 'TIFF']
outputbuttons = CW_BGROUP(skymapbase, outputnames, UVALUE = 'OUTPUTBUTTON', $
                    /ROW, /EXCLUSIVE, /RETURN_NAME, SET_VALUE = 0, $
                    UNAME = 'outputbutton', LABEL_LEFT = 'Output type')
screen = 1B
postscript = 0B
jpg = 0B
tiff = 0B

savefilewidget = CW_FIELD(skymapbase, TITLE = 'Filename:', UVALUE = $
                        'SAVEFILE', VALUE = 'skymap', XSIZE = 20)

IF mainuval.skystate.use THEN BEGIN
   savefile = STRTRIM(mainuval.skystate.savefile,2)
   WIDGET_CONTROL, savefilewidget, SET_VALUE = savefile
ENDIF ELSE savefile = ''

IF NOT MAINUVAL.skystate.use THEN BEGIN
    mainuval.skystate.latlon = latlon
    mainuval.skystate.alltimes = alltimes
    mainuval.skystate.animate = skyanim
    mainuval.skystate.cubetimes = cubetimes
    mainuval.skystate.screen = screen
    mainuval.skystate.postscript = postscript
    mainuval.skystate.jpg = jpg
    mainuval.skystate.tiff = tiff
    mainuval.skystate.savefile = savefile
ENDIF

IF mainuval.skystate.use THEN BEGIN
    WIDGET_CONTROL, redmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.skystate.spec_max[0]),2)
    WIDGET_CONTROL, greenmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.skystate.spec_max[1]),2)
    WIDGET_CONTROL, bluemaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.skystate.spec_max[2]),2)
    WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.skystate.ntime),2)
    WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.skystate.maptime),2)

    IF mainuval.skystate.animate THEN BEGIN
       WIDGET_CONTROL, animyes, /SET_BUTTON
       skyanim = 1B
    ENDIF
    IF mainuval.skystate.alltimes THEN BEGIN
       WIDGET_CONTROL, alltimesbutton, /SET_BUTTON
       alltimes = 1B
    ENDIF
    IF mainuval.skystate.cubetimes THEN BEGIN
       WIDGET_CONTROL, cubeyes, /SET_BUTTON
       cubetimes = 1B
    ENDIF
    IF mainuval.skystate.screen THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
       screen = 1B
    ENDIF
    IF mainuval.skystate.postscript THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 1
       postscript = 1B
       savefile = mainuval.skystate.savefile
    ENDIF
    IF mainuval.skystate.jpg THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 2
       jpg = 1B
       savefile = mainuval.skystate.savefile
    ENDIF
    IF mainuval.skystate.tiff THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 3
       tiff = 1B
       savefile = mainuval.skystate.savefile
    ENDIF
    IF NOT mainuval.skystate.latlon THEN BEGIN
       latlon = 0B
       WIDGET_CONTROL, hidelatlonbutton, /SET_BUTTON
    ENDIF
ENDIF


buttonbase = WIDGET_BASE(skymapbase, /ROW)
helpbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Help', $
                           UVALUE = 'HELPBUTTON')
cancelbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Cancel', $
                             UVALUE = 'CANCELBUTTON')
;; Clicking the done button on this widget results in all 
;; values being read from the widget and passed to
;; DIRECT_SKYMAP.
skymapdonebutton = WIDGET_BUTTON(buttonbase, VALUE = 'Make it so', $
                                 UVALUE = 'DONEBUTTON', SENSITIVE=0)

WIDGET_CONTROL, skymapbase, SET_UVALUE = skymapuval
WIDGET_CONTROL, skymapbase, /REALIZE, GROUP = geometer_base, /MODAL

IF mainuval.skystate.cubetimes THEN $
       WIDGET_CONTROL, skymapdonebutton, SENSITIVE = 1

;; Check this widget to see if the Done button is pressed.
;; Once the done button is pressed, read the values from the
;; various inputs in this widget, destroy the widget, and then
;; send these inputs to the skymap routine.

done_with_skymapinput = 0B
spec_bounds = [0,1,2,3,4,5]

REPEAT BEGIN
   skymapevent = WIDGET_EVENT()
   CASE skymapevent.top OF
      skymapbase:BEGIN
         WIDGET_CONTROL, skymapevent.id, GET_UVALUE = ev_uval
         CASE ev_uval OF
            'HELPBUTTON': XDISPLAYFILE, mainuval.geometerpath+$
                'directskymaphelp.txt', DONE_BUTTON = 'Done', $
                                        GROUP = skymapbase, /MODAL
            'CANCELBUTTON': BEGIN
               WIDGET_CONTROL, skymapevent.top, /DESTROY
               RETURN
            END
            'DONEBUTTON':BEGIN
               WIDGET_CONTROL, raminwidget, GET_VALUE = ramin
               ramin = FLOAT(ramin[0])
               mainuval.skystate.ramin = ramin
               WIDGET_CONTROL, ramaxwidget, GET_VALUE = ramax
               ramax = FLOAT(ramax[0])
               mainuval.skystate.ramax = ramax
               WIDGET_CONTROL, deminwidget, GET_VALUE = demin
               demin = FLOAT(demin[0])
                    mainuval.skystate.decmin = demin
               WIDGET_CONTROL, demaxwidget, GET_VALUE = demax
               demax = FLOAT(demax[0])
               mainuval.skystate.decmax = demax

               ;; Read out the color selections:
               WIDGET_CONTROL, redwidget, GET_VALUE = red
               redlims = STRSPLIT(red,',',/EXTRACT)
               spec_bounds[0] = FIX(redlims[0])
               spec_bounds[1] = FIX(redlims[1])
               WIDGET_CONTROL, redmaxwidget, GET_VALUE = redmaxstring
               redmax = FLOAT(redmaxstring[0])
               
               WIDGET_CONTROL, greenwidget, GET_VALUE = green
               greenlims = STRSPLIT(green,',',/EXTRACT)
               spec_bounds[2] = FIX(greenlims[0])
               spec_bounds[3] = FIX(greenlims[1])
               WIDGET_CONTROL, greenmaxwidget, GET_VALUE = greenmaxstring
               greenmax = FLOAT(greenmaxstring[0])

               WIDGET_CONTROL, bluewidget, GET_VALUE = blue
               bluelims = STRSPLIT(blue,',',/EXTRACT)
               spec_bounds[4] = FIX(bluelims[0])
               spec_bounds[5] = FIX(bluelims[1])
               WIDGET_CONTROL, bluemaxwidget, GET_VALUE = bluemaxstring
               bluemax = FLOAT(bluemaxstring[0])

               mainuval.skystate.spec_bounds = spec_bounds
               mainuval.skystate.spec_max = [redmax, greenmax, bluemax]

               WIDGET_CONTROL, ntimewidget, GET_VALUE = ntimestring
               ntimeplot = FIX(FLOAT(ntimestring[0]))
               mainuval.skystate.ntime = ntimeplot

               WIDGET_CONTROL, maptimewidget, GET_VALUE = maptimestring
               maptime = STRTRIM(maptimestring[0],2)
               mainuval.skystate.maptime = maptime

               WIDGET_CONTROL, savefilewidget, GET_VALUE = savefile
               mainuval.skystate.savefile = savefile

               done_with_skymapinput = 1B
               mainuval.skystate.use = 1B
            END
            'PLOTBUTTON': BEGIN
               WIDGET_CONTROL, plotrowselect, GET_VALUE = rowselect
               rowselect = FIX(FLOAT(rowselect[0]))
               nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
               IF rowselect GE nrows THEN BEGIN
                  WIDGET_CONTROL, plotrowselect, SET_VALUE = $
                                  STRTRIM(STRING(nrows-1),2)
                  rowselect = nrows - 1
               ENDIF
               WIDGET_CONTROL, plottimeselect, GET_VALUE = timeselect
               timeselect = FIX(FLOAT(timeselect[0]))
               nrecords = N_ELEMENTS(cube.et)
               IF timeselect GE nrecords THEN BEGIN
                  WIDGET_CONTROL, plottimeselect, SET_VALUE = $
                                  STRTRIM(STRING(nrecords-1), 2)
                  timeselect = nrecords-1
               ENDIF
               PLOT, cube[timeselect].data[rowselect, *], XTITLE = $
                     'Channel Number', YTITLE = 'Data Cube Value', $
                          XRANGE = [0, nspec], $
                          TITLE = 'Spectrum for: Record '+$
                          STRTRIM(STRING(timeselect),2)+'. Row '+$
                          STRTRIM(STRING(rowselect),2)+'.', $
                          YRANGE = $
                            [0., MAX(cube[timeselect].data[rowselect,*])]
            END
            'DATAINFO': BEGIN
               ;; Create an informational widget displaying summary
               ;; info on the loaded data cube and on the currently
               ;; selected row and record. 
               ;; Get selected row:
               WIDGET_CONTROL, plotrowselect, GET_VALUE = rowselect
               rowselect = FIX(FLOAT(rowselect[0])) > 0
               nrows = N_ELEMENTS(cube[0].data[*,0]) 
               IF rowselect GE nrows THEN BEGIN
                  WIDGET_CONTROL, plotrowselect, SET_VALUE = $
                                  STRTRIM(STRING(nrows-1),2)
                  rowselect = nrows - 1
               ENDIF
               
               ;; Get selected record:
               WIDGET_CONTROL, plottimeselect, GET_VALUE = timeselect
               timeselect = FIX(FLOAT(timeselect[0])) > 0
               nrecords = N_ELEMENTS(cube.et)
               IF timeselect GE nrecords THEN BEGIN
                  WIDGET_CONTROL, timeselect, SET_VALUE = $
                                  STRTRIM(STRING(nrecords-1), 2)
                  timeselect = nrecords-1
               ENDIF
               
               ;; Get the selected color ranges:
               WIDGET_CONTROL, redwidget, GET_VALUE = red
               redlims = STRSPLIT(red,',',/EXTRACT)
               spec_bounds[0] = FIX(redlims[0])
               spec_bounds[1] = FIX(redlims[1])
               
               WIDGET_CONTROL, greenwidget, GET_VALUE = green
               greenlims = STRSPLIT(green,',',/EXTRACT)
               spec_bounds[2] = FIX(greenlims[0])
               spec_bounds[3] = FIX(greenlims[1])
               
               WIDGET_CONTROL, bluewidget, GET_VALUE = blue
               bluelims = STRSPLIT(blue,',',/EXTRACT)
               spec_bounds[4] = FIX(bluelims[0])
               spec_bounds[5] = FIX(bluelims[1])
               
               row_moment = STRING(MOMENT(cube[*].data[$
                            rowselect,0:nspec-1]))
               row_max_min = STRTRIM(STRING(MAX(cube[*].data[$
                             rowselect,0:nspec-1])),2)+' '+STRTRIM(STRING($
                             MIN(cube[*].data[rowselect,0:nspec-1])),2)
               
               record_moment = STRING(MOMENT(cube[timeselect].data[*,*]))
               rec_max_min = STRTRIM(STRING(MAX(cube[timeselect].data[$
                             0:nrows-1,0:nspec-1])),2)+' '+$
                             STRTRIM(STRING($
                             MIN(cube[timeselect].data[*,*])),2)
               
               cube_moment = STRING(MOMENT(cube[*].data[$
                             0:nrows-1,0:nspec-1]))
               cube_max_min = STRTRIM(STRING(MAX(cube.data)),2)+' '+$
                              STRTRIM(STRING(MIN(cube.data[$
                              0:nrows-1,nspec-1])),2)
               red_moment = STRING(MOMENT(cube[*].data[*,spec_bounds[0]:$
                                                       spec_bounds[1]]))
               red_max_min = STRTRIM(STRING(TOTAL(cube[*].data[$
                             *,spec_bounds[0]:spec_bounds[1]])),2)+$
                             ' '+STRTRIM(STRING(MIN(cube[*].data[$
                             *,spec_bounds[0]:spec_bounds[1]])),2)
               green_moment = STRING(MOMENT(cube[*].data[$
                              *,spec_bounds[2]:spec_bounds[3]]))
               green_max_min = STRTRIM(STRING(TOTAL(cube[*].data[$
                               *,spec_bounds[2]:spec_bounds[3]])),2)+$
                               ' '+STRTRIM(STRING(MIN(cube[*].data[$
                               *,spec_bounds[2]:spec_bounds[3]])),2)
               blue_moment = STRING(MOMENT(cube[*].data[$
                             *,spec_bounds[4]:spec_bounds[5]]))
               blue_max_min = STRTRIM(STRING(TOTAL(cube[*].data[$
                              *,spec_bounds[4]:spec_bounds[5]])),2)+$
                              ' '+STRTRIM(STRING(MIN(cube[*].data[$
                              *,spec_bounds[4]:spec_bounds[5]])),2)
               
               ;; Put information in a string array for display
               ;; in the widget.
               line0 = '                          Mean    '+$
                       '       Std. Dev.        Skewness'+$
                       '       Kurtosis    Maximum  Minimum'
               line1 = 'Red statistics:   '+red_moment[0]+$
                       red_moment[1]+red_moment[2]+$
                       red_moment[3]+'  '+red_max_min
               line2 = 'Green statistics: '+green_moment[0]+$
                       green_moment[1]+green_moment[2]+$
                       green_moment[3]+'  '+green_max_min
               line3 = 'Blue statistics:  '+blue_moment[0]+$
                       blue_moment[1]+blue_moment[2]+$
                       blue_moment[3]+'  '+blue_max_min
               line4 = 'Row '+STRTRIM(STRING(rowselect),2)+$
                       ' statistics:   '+row_moment[0]+$
                       row_moment[1]+row_moment[2]+$
                       row_moment[3]+'  '+row_max_min
               line5 = 'Record '+STRTRIM(STRING(timeselect),2)+$
                       ' statistics:'+record_moment[0]+$
                       record_moment[1]+record_moment[2]+$
                       record_moment[3]+'  '+rec_max_min
               line6 = 'Cube statistics:  '+cube_moment[0]+$
                       cube_moment[1]+cube_moment[2]+$
                       cube_moment[3]+'  '+cube_max_min
               buffer = [line0, line1, line2, line3, line4, line5, line6]
               
               junk = DIALOG_MESSAGE(buffer, TITLE = $
                                     'Data Cube Statistics',$
                                     /CENTER, /INFORMATION)
            END
            'ALLTIMES': BEGIN
               alltimes = 1B
               mainuval.skystate.alltimes = 1B
            END
            'ONETIME': BEGIN 
               alltimes = 0B
               mainuval.skystate.alltimes = 0B
            END
            'OUTPUTBUTTON': BEGIN
               WIDGET_CONTROL, outputbuttons, GET_VALUE = outputtype
               CASE outputtype OF
                  0: BEGIN
                     screen = 1B
                     postscript = 0B
                     jpg = 0B
                     tiff = 0B
                     mainuval.skystate.screen = 1B
                     mainuval.skystate.postscript = 0B
                     mainuval.skystate.jpg = 0B
                     mainuval.skystate.tiff = 0B
                  END
                  1: BEGIN
                     screen = 0B
                     postscript = 1B
                     jpg = 0B
                     tiff = 0B
                     mainuval.skystate.screen = 0B
                     mainuval.skystate.postscript = 1B
                     mainuval.skystate.jpg = 0B
                     mainuval.skystate.tiff = 0B
                     mainuval.skystate.savefile = savefile
                  END
                  2: BEGIN
                     screen = 0B
                     jpg = 1B
                     postscript = 0B
                     tiff = 0B
                     mainuval.skystate.screen = 0B
                     mainuval.skystate.postscript = 0B
                     mainuval.skystate.jpg = 1B
                     mainuval.skystate.tiff = 0B
                     mainuval.skystate.savefile = savefile
                  END
                  3: BEGIN
                     screen = 0B
                     jpg = 0B
                     postscript = 0B
                     tiff = 1B
                     mainuval.skystate.screen = 0B
                     mainuval.skystate.postscript = 0B
                     mainuval.skystate.jpg = 0B
                     mainuval.skystate.tiff = 1B
                     mainuval.skystate.savefile = savefile
                  END
               ENDCASE
            END
            'ANIMYES': BEGIN
               skyanim = 1B
               mainuval.skystate.animate = 1B
               screen = 1B
               postscript = 0B
               jpg = 0B
               tiff = 0B
            END
            'ANIMNO': BEGIN
               skyanim = 0B
               mainuval.skystate.animate = 0B
            END
            'LATLONYES': BEGIN 
               latlon = 1B
               mainuval.skystate.latlon = 1B
            END
            'LATLONNO': BEGIN 
               latlon = 0B
               mainuval.skystate.latlon = 0B
            END
            'MAPTIMETEXT': BEGIN
               WIDGET_CONTROL, maptimewidget, GET_VALUE = $
                               maptimestring
               CSPICE_UTC2ET, maptimestring[0], maptimeet
               CSPICE_ET2UTC, maptimeet, 'ISOD', 3L, maptimestring
               WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                               maptimestring
            END
            'CUBEYES': BEGIN
               limbtime = cube[0].et

               ntimes = N_ELEMENTS(cube.et)
               IF NOT mainuval.skystate.use THEN $
               WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                               STRING(STRTRIM(ntimes),2)

               BODY_LIMB, target, 'CASSINI', limbtime, 90, $
                          ra_limb, dec_limb, limberror

               ;; Allow user to continue
               WIDGET_CONTROL, skymapdonebutton, SENSITIVE = 1

               IF mainuval.skyzoom EQ 2 THEN BEGIN
                  ra_min = MAX(ra_limb)
                  ra_max = MIN(ra_limb)
                  ra_center = ra_min-.5*ABS(ra_max-ra_min)
                  delta_ra = zoomscalefactor*ABS(ra_max-ra_min) < 360.
                  dec_min = MIN(dec_limb)
                  dec_max = MAX(dec_limb)
                  dec_center = dec_min+.5*(dec_max-dec_min)
                  delta_dec = zoomscalefactor*(dec_max-dec_min) < 180.
                        
                  ra_min = (ra_center+.5*delta_ra) > 0.
                  ra_max = (ra_min-delta_ra) < 360.
                  dec_min = (dec_center-.5*delta_dec) > (-90.)
                  dec_max = (dec_min+delta_dec) < 90.
                  IF NOT mainuval.skystate.use THEN BEGIN
                     WIDGET_CONTROL, raminwidget, SET_VALUE = $
                                     STRTRIM(STRING(ra_min),2)
                     WIDGET_CONTROL, ramaxwidget, SET_VALUE = $
                                     STRTRIM(STRING(ra_max),2)
                     WIDGET_CONTROL, deminwidget, SET_VALUE = $
                                     STRTRIM(STRING(dec_min),2)
                     WIDGET_CONTROL, demaxwidget, SET_VALUE = $
                                     STRTRIM(STRING(dec_max), 2)
                  ENDIF
               ENDIF
               cubetimes = 1B
               mainuval.skystate.cubetimes = 1B
            END
            'CUBENO': BEGIN
               IF N_ELEMENTS(data) NE 0 THEN BEGIN
               limbtime = data[0].et

               ntimes = N_ELEMENTS(data.ra_full[0])
               WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                               STRING(STRTRIM(ntimes),2)
               BODY_LIMB, target, 'CASSINI', limbtime, 90, $
                          ra_limb, dec_limb, limberror
               ;; Allow user to continue
               WIDGET_CONTROL, skymapdonebutton, SENSITIVE = 1

               IF mainuval.skyzoom EQ 2 THEN BEGIN
                  ra_min = MAX(ra_limb)
                  ra_max = MIN(ra_limb)
                  ra_center = ra_min-.5*ABS(ra_max-ra_min)
                  delta_ra = zoomscalefactor*ABS(ra_max-ra_min) < 360.
                  dec_min = MIN(dec_limb)
                  dec_max = MAX(dec_limb)
                  dec_center = dec_min+.5*(dec_max-dec_min)
                  delta_dec = zoomscalefactor*(dec_max-dec_min) < 180.
                  
                  ra_min = (ra_center+.5*delta_ra) > 0.
                  ra_max = (ra_min-delta_ra) < 360.
                  dec_min = (dec_center-.5*delta_dec) > (-90.)
                  dec_max = (dec_min+delta_dec) < 90.
                  IF NOT mainuval.skystate.use THEN BEGIN
                     WIDGET_CONTROL, raminwidget, SET_VALUE = $
                                     STRTRIM(STRING(ra_min),2)
                     WIDGET_CONTROL, ramaxwidget, SET_VALUE = $
                                     STRTRIM(STRING(ra_max),2)
                     WIDGET_CONTROL, deminwidget, SET_VALUE = $
                                     STRTRIM(STRING(dec_min),2)
                     WIDGET_CONTROL, demaxwidget, SET_VALUE = $
                                     STRTRIM(STRING(dec_max), 2)
                  ENDIF
               ENDIF
               cubetimes = 0B
               mainuval.skystate.cubetimes = 0B
            ENDIF               ; check that data is defined
            END
            ELSE:
         ENDCASE
      END
      ELSE: ;; Do nothing
   ENDCASE
ENDREP UNTIL done_with_skymapinput
WIDGET_CONTROL, skymapbase, /DESTROY

IF cubetimes THEN ntimes = N_ELEMENTS(cube.et) ELSE $
   ntimes = N_ELEMENTS(data.ra_full[0])

IF skyanim THEN XINTERANIMATE, SET = [800, 400, ntimeplot] 

;; Identify the time to make the map from the user entry:
IF NOT cubetimes THEN BEGIN
;    WIDGET_CONTROL, mainuval.maptimetext, GET_VALUE = maptime
;    maptime = maptime[0]
;print,cubetimes
;print,maptime
    CSPICE_UTC2ET, maptime, mapet
    junk = MIN(ABS(mapet-data.et),etloc)
    CSPICE_ET2UTC, data[etloc].et, mainuval.tformat, 3L, maptime
    mainuval.skystate.maptime = maptime
    WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                    'Actual time for map: ' + maptime
ENDIF ELSE BEGIN
   CSPICE_UTC2ET, maptime, mapet
   junk = MIN(ABS(mapet-cube.et), etloc)
   CSPICE_ET2UTC, cube[etloc].et, mainuval.tformat, 3L, maptime
   mainuval.skystate.maptime = maptime
ENDELSE

;; Get star limiting magnitude:
WIDGET_CONTROL, mainuval.mapstarlimtext, GET_VALUE = starlim
starlim = FLOAT(starlim[0])

;; Determine which set of times to use and store in vectors
;; to pass to DIRECT_SKYMAP.

dtime_ix = FLOAT(ntimes)/(ntimeplot-1)
plot_ets = DBLARR(ntimeplot)
plot_utcs = STRARR(ntimeplot)

IF cubetimes THEN BEGIN
    IF mainuval.cubeloaded THEN BEGIN
        cube_skymap = cube
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = cube[gettime_ix].et
            CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                           3L, utc0
            plot_utcs[time_ix] = utc0
        ENDFOR
    ENDIF ELSE BEGIN
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = data[gettime_ix].et
            plot_utcs[time_ix] = data[gettime_ix].utc
        ENDFOR
        WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                        'Cube not loaded.'
        WIDGET_CONTROL, skymapbase, /DESTROY
        RETURN
    ENDELSE
ENDIF ELSE BEGIN
    ;; Create a dummy variable to take the place of the cube
    ;; data to pass to DIRECT_SKYMAP since cube data may
    ;; not be loaded if we're here.
    cube_skymap = {et:FLTARR(1)+1E20} 
    ;; Give it a ridiculous value so
    ;; DIRECT_SKYMAP will know it's not real.

    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        gettime_ix = time_ix*dtime_ix < (ntimes-1)
        plot_ets[time_ix] = data[gettime_ix].et
        CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                       3L, utc0
        plot_utcs[time_ix] = utc0
    ENDFOR
ENDELSE

;; Now make the call(s) to DIRECT_SKYMAP
IF skyanim THEN BEGIN
    maptimedummy = ''
    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        data_ix = time_ix*dtime_ix < (ntimes-1)
        DIRECT_SKYMAP, data, mainuval, data_ix, time_ix, plot_ets, $
                       plot_utcs, cube_skymap, spec_bounds, maptimedummy, $
                       RA_RANGE = [ramin, ramax], $
                       DEC_RANGE = [demin, demax], $
                       COLORMAX = [redmax, greenmax, bluemax], $
                       ALLTIMES = alltimes, STARS = starlim, $ 
                       LATLON = latlon, $
                       SCREEN = 1, POSTSCRIPT = 0, xsize = 800, ysize = 400, $
                       JPG = 0, TIFF = 0, SAVEFILE = savefile
        XINTERANIMATE, FRAME = time_ix, WINDOW = !D.WINDOW
    ENDFOR
    XINTERANIMATE
ENDIF ELSE $
  DIRECT_SKYMAP, data, mainuval, etloc, etloc, plot_ets, $
                 plot_utcs, cube_skymap, spec_bounds, maptime, $
                 RA_RANGE = [ramin, ramax], $
                 DEC_RANGE = [demin, demax], $
                 COLORMAX = [redmax, greenmax, bluemax], $
                 ALLTIMES = alltimes, STARS = starlim, $
                 LATLON = latlon, $
                 SCREEN = screen, POSTSCRIPT = postscript, $
                 JPG = jpg, TIFF = tiff, SAVEFILE = savefile, $
                 XSIZE = 1000, YSIZE = 500


END
