;; File Version: 1.3
;; Last Modified: June 16, 2018
;; By: JEC

PRO CALL_TARGET, mainuval, data, cube, geometer_base

COMPILE_OPT idl2

;; Procedure must be called by GEOMETER. Builds a temporary widget to 
;; collect user preferences for the DIRECT_TARGET plot. Based on 
;; DIRECT_SKYMAP.
;; Josh Colwell
;; July 31, 2006
;; August 21, 2006: Multiple user options enabled. Base checkout complete.
;; August 24, 2006: edited by AWD to allow target mapping with or without 
;;   data cube input 
;; December 6, 2006: allow redrawing grid over data.
;; December 7, 2006: Postscript output option.
;; January 10, 2007: Begin implementing graphic state vector.
;; January 16, 2007: Debugged and expanded graphic state vector.
;; January 17, 2007: added maptime option. Alex Ware DeWolfe.
;; March 1, 2007: Modify for GEOMETER version of Data Cube.
;; April 3, 2007: Fixed animate bug.
;; May 25, 2007: Handle F Ring target.
;; October 9, 2007: JPEG output option.
;; December 18, 2007: Updated to allow multiple output file formats. 
;; January 19, 2010: Add the FORCECOORDS option. This enables the user
;; to force Direct_Target to use the entered RA/Dec ranges exactly,
;; without preserving perspective. JEC.
;; July 11, 2017: Add ability to overplot ISS NAC and WAC FOVs in addition to 
;; the selected FOV channel. JEC.
;; June 16, 2018: Commented out the re-use of previous values of some drawing settings.
;; While potentially a time-save, it made it awkward to change the zoom. Updated to
;; work with new zoom feature in geometer.

;; Create a window to get plotting preferences.
targetbase = WIDGET_BASE(/COLUMN, TITLE = 'Direct Target Options')

;; Get the target body limb limits
IF mainuval.cubeloaded THEN limbtime = cube[0].et ELSE $
  limbtime = data[0].et
targetid = WIDGET_INFO(mainuval.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, mainuval.targetdrop, GET_VALUE = $
                targetdropvalue
target = STRUPCASE(targetdropvalue[targetid])
IF STRUPCASE(target) EQ 'RINGS' THEN target = 'SATURN'
IF STRUPCASE(target) EQ 'F RING' THEN target = 'SATURN'
;; Get the camera rotation matrix used to create perspective view:
CSPICE_SPKPOS, target, limbtime, 'J2000', 'LT', 'CASSINI', targpos, ltime
CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, ratarg, dectarg
CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, 3L, 2L, 3L, cam_mat

BODY_LIMB, target, 'CASSINI', limbtime, 90, ra_limb, dec_limb, limberror
RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
         u_limb, v_limb, behind, cam_mat

;; Get the zoom value: 
;IF mainuval.skyzoom EQ 2 THEN zoomscalefactor = 1.5 ELSE zoomscalefactor = 4.
zoomscalefactor = mainuval.skyzoom

u_min = MIN(u_limb)
u_max = MAX(u_limb)
u_center = MEAN([u_min,u_max])
delta_u = zoomscalefactor*ABS(u_max-u_min)
v_min = MIN(v_limb)
v_max = MAX(v_limb)
v_center = MEAN([v_min, v_max])
delta_v = zoomscalefactor*ABS(v_max-v_min) 

u_min = u_center+.5*delta_u
u_max = u_min-delta_u
v_min = v_center-.5*delta_v
v_max = v_min+delta_v

UV_2_RD, u_min, v_min, ra_min, dec_min, cam_mat
UV_2_RD, u_max, v_max, ra_max, dec_max, cam_mat

rabase = WIDGET_BASE(targetbase, /ROW)
raminwidget = CW_FIELD(rabase, TITLE = 'Min. RA (deg): ',$
                       UVALUE = 'RAMINTEXT', VALUE = $
                       STRTRIM(STRING(ra_min),2), XSIZE = 12)
ramaxwidget = CW_FIELD(rabase, TITLE = 'Max. RA (deg): ',$
                       UVALUE = 'RAMAXTEXT', VALUE = $
                       STRTRIM(STRING(ra_max),2), XSIZE = 12)
decbase = WIDGET_BASE(targetbase, /ROW)
deminwidget = CW_FIELD(decbase, TITLE = 'Min. Dec (deg): ',$
                       UVALUE = 'DEMINTEXT', VALUE= $
                       STRTRIM(STRING(dec_min),2), XSIZE = 12)
demaxwidget = CW_FIELD(decbase, TITLE = 'Max. Dec (deg): ',$
                       UVALUE = 'DEMAXTEXT', VALUE = $
                       STRTRIM(STRING(dec_max),2), XSIZE = 12)

gridrowbase = WIDGET_BASE(targetbase, /ROW)
label = WIDGET_LABEL(gridrowbase, VALUE = 'Show RA/Dec Grid:')
griddrawbase = WIDGET_BASE(gridrowbase, /ROW, /EXCLUSIVE)
gridyesbutton = WIDGET_BUTTON(griddrawbase, VALUE = $
                              'Yes', UVALUE = 'GRIDYES')
gridnobutton = WIDGET_BUTTON(griddrawbase, VALUE = 'No',  UVALUE = 'GRIDNO')

WIDGET_CONTROL, gridyesbutton, /SET_BUTTON

grid = 1B

label = WIDGET_LABEL(gridrowbase, VALUE = 'Exact Coords:')
forcebase = WIDGET_BASE(gridrowbase, /ROW, /EXCLUSIVE)
forceyesbutton = WIDGET_BUTTON(forcebase, VALUE = 'Yes', UVALUE = 'FORCEYES')
forcenobutton  = WIDGET_BUTTON(forcebase, VALUE = 'No', UVALUE = 'FORCENO')

WIDGET_CONTROL, forcenobutton, /SET_BUTTON
forcecoords = 0B


;; Check to see if values from a previous run of target exist, and if so,
;; use those values.
;; Commenting this out. While it saves time sometimes, it also makes it impossible to change the 
;; Zoom. JEC. 6/16/18
;IF mainuval.targstate.use THEN BEGIN
;    IF NOT mainuval.targstate.showgrid THEN BEGIN
;        grid = 0B
;        WIDGET_CONTROL, gridnobutton, /SET_BUTTON
;    ENDIF
;    WIDGET_CONTROL, raminwidget, SET_VALUE = $
;                    STRTRIM(STRING(mainuval.targstate.ramin),2)
;    WIDGET_CONTROL, ramaxwidget, SET_VALUE = $
;                    STRTRIM(STRING(mainuval.targstate.ramax),2)
;    WIDGET_CONTROL, deminwidget, SET_VALUE = $
;                    STRTRIM(STRING(mainuval.targstate.decmin),2)
;    WIDGET_CONTROL, demaxwidget, SET_VALUE = $
;                    STRTRIM(STRING(mainuval.targstate.decmax),2)
;ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
    specminred = '0'
    specmaxred = '0'
    specmingreen = '0'
    specmaxgreen = '0'
    specminblue = '0'
    specmaxblue = '0'
ENDIF ELSE BEGIN
    nspec = N_ELEMENTS(WHERE(cube[0].data[0,*] NE -1.))
    IF mainuval.targstate.use THEN BEGIN
        specminred = STRTRIM(STRING(mainuval.targstate.spec_bounds[0]),2)
        specmaxred = STRTRIM(STRING(mainuval.targstate.spec_bounds[1]),2)
        specmingreen = STRTRIM(STRING(mainuval.targstate.spec_bounds[2]),2)
        specmaxgreen = STRTRIM(STRING(mainuval.targstate.spec_bounds[3]),2)
        specminblue = STRTRIM(STRING(mainuval.targstate.spec_bounds[4]),2)
        specmaxblue = STRTRIM(STRING(mainuval.targstate.spec_bounds[5]),2)
    ENDIF ELSE BEGIN
        specminred = '0'
        specmaxred = STRTRIM(STRING(nspec/3),2)
        specmingreen = STRTRIM(STRING(nspec/3+1),2)
        specmaxgreen = STRTRIM(STRING(nspec-nspec/3),2)
        specminblue = STRTRIM(STRING(nspec-nspec/3+1),2)
        specmaxblue = STRTRIM(STRING(nspec-1),2)
    ENDELSE
ENDELSE

plotspecbase = WIDGET_BASE(targetbase, /ROW, /FRAME)
plotrowselect = CW_FIELD(plotspecbase, TITLE = 'Plot Row: ', $
                         UVALUE = 'PLOTROW', VALUE = '0', XSIZE = 3)
plottimeselect = CW_FIELD(plotspecbase, TITLE = 'Plot Index: ', $
                          UVALUE = 'PLOTTIME', VALUE = '0', XSIZE =3)
plotbutton = WIDGET_BUTTON(plotspecbase, VALUE = 'Plot Spectrum', $
                           UVALUE = 'PLOTBUTTON')
infobutton = WIDGET_BUTTON(plotspecbase, VALUE = 'Data Info', UVALUE = $
                           'DATAINFO')
IF NOT mainuval.cubeloaded THEN WIDGET_CONTROL, plotspecbase, SENSITIVE = 0

redbase = WIDGET_BASE(targetbase, /ROW)
redwidget = CW_FIELD(redbase, TITLE = 'Red channels (lo,hi): ',$
                     UVALUE = 'REDTEXT', VALUE = specminred+','+specmaxred, $
                     XSIZE = 8)
redmaxwidget = CW_FIELD(redbase, TITLE = 'Red maximum: ', UVALUE = $
                        'REDMAX', VALUE = '0', XSIZE = 6)

greenbase = WIDGET_BASE(targetbase, /ROW)
greenwidget = CW_FIELD(greenbase, TITLE = 'Green channels (lo,hi): ',$
                       UVALUE = 'GREENTEXT', VALUE = specmingreen+','+$
                       specmaxgreen, XSIZE = 8)
greenmaxwidget = CW_FIELD(greenbase, TITLE = 'Green maximum: ', UVALUE = $
                        'GREENMAX', VALUE = '0', XSIZE = 6)

bluebase = WIDGET_BASE(targetbase, /ROW)
bluewidget = CW_FIELD(bluebase, TITLE = 'Blue channels (lo,hi): ',$
                      UVALUE = 'BLUETEXT', VALUE = specminblue+','+$
                      specmaxblue, XSIZE = 8)
bluemaxwidget = CW_FIELD(bluebase, TITLE = 'Blue maximum: ', UVALUE = $
                        'BLUEMAX', VALUE = '0', XSIZE = 6)

;; Have A Series of Yes/No buttons for various drawing options.
optionsbase1 = WIDGET_BASE(targetbase, /ROW)
baseforframe1 = WIDGET_BASE(optionsbase1, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe1, VALUE = 'Show Shadow: ')
shadowbase = WIDGET_BASE(baseforframe1, /ROW, /EXCLUSIVE)
showshadowbutton = WIDGET_BUTTON(shadowbase, VALUE = 'Yes', $
                                 UVALUE = 'SHADOWYES')
hideshadowbutton = WIDGET_BUTTON(shadowbase, VALUE = 'No', $
                                 UVALUE = 'SHADOWNO')
WIDGET_CONTROL, showshadowbutton, /SET_BUTTON
shadow = 1B

baseforframe2 = WIDGET_BASE(optionsbase1, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe2, VALUE = 'Show Lat/Lon: ')
latlonbase = WIDGET_BASE(baseforframe2, /ROW, /EXCLUSIVE)
showlatlonbutton = WIDGET_BUTTON(latlonbase, VALUE = 'Yes', $
                                 UVALUE = 'LATLONYES')
hidelatlonbutton = WIDGET_BUTTON(latlonbase, VALUE = 'No', $
                                 UVALUE = 'LATLONNO')
WIDGET_CONTROL, showlatlonbutton, /SET_BUTTON
latlon = 1B

baseforiss = WIDGET_BASE(targetbase, /ROW, /FRAME)
label = WIDGET_LABEL(baseforiss, VALUE = 'Draw ISS NAC: ')
baseforissnac = WIDGET_BASE(baseforiss, /ROW, /EXCLUSIVE)
drawissnac = WIDGET_BUTTON(baseforissnac, VALUE = 'Yes', UVALUE = 'ISSNACYES')
hideissnac = WIDGET_BUTTON(baseforissnac, VALUE = 'No', UVALUE = 'ISSNACNO')
WIDGET_CONTROL, hideissnac, /SET_BUTTON
issnac = 0B
label = WIDGET_LABEL(baseforiss, VALUE = 'Draw ISS WAC: ')
baseforisswac = WIDGET_BASE(baseforiss, /ROW, /EXCLUSIVE)
drawisswac = WIDGET_BUTTON(baseforisswac, VALUE = 'Yes', UVALUE = 'ISSWACYES')
hideisswac = WIDGET_BUTTON(baseforisswac, VALUE = 'No', UVALUE = 'ISSWACNO')
WIDGET_CONTROL, hideisswac, /SET_BUTTON
isswac = 0B

baseforframe3 = WIDGET_BASE(targetbase, /ROW, /FRAME)
label = WIDGET_LABEL(baseforframe3, VALUE = 'Lat/Lon over pixels:')
gridontopbase = WIDGET_BASE(baseforframe3, /ROW, /EXCLUSIVE)
gridontop = WIDGET_BUTTON(gridontopbase, VALUE='Yes', $
                          UVALUE = 'GRIDTOP')
gridonbottom = WIDGET_BUTTON(gridontopbase, VALUE = 'No', $
                             UVALUE = 'GRIDBOTTOM')
WIDGET_CONTROL, gridontop, /SET_BUTTON
gridtop = 1B

;; Allow user to select subset of available times:
timeselectbase = WIDGET_BASE(targetbase, /ROW, /FRAME)
ntimewidget = CW_FIELD(timeselectbase, TITLE = 'Number of time steps to plot', $
                       VALUE = '1', UVALUE = 'TIMETEXT', XSIZE = 8)

;; Allow user to show all FOV pointings at once (but a single
;; background).
timesbase = WIDGET_BASE(targetbase, /ROW, /EXCLUSIVE)
alltimesbutton = WIDGET_BUTTON(timesbase, VALUE = $
                               'Show all times', $
                               UVALUE = 'ALLTIMES')
onetimebutton = WIDGET_BUTTON(timesbase, VALUE = $
                              'Show selected time', $
                              UVALUE = 'ONETIME')
WIDGET_CONTROL, onetimebutton, /SET_BUTTON
alltimes = 0B

;; Allow user to select an animation:
animbase = WIDGET_BASE(targetbase, /ROW, /EXCLUSIVE)
animyes = WIDGET_BUTTON(animbase, VALUE = 'Animate', $
                        UVALUE = 'ANIMYES')
animno = WIDGET_BUTTON(animbase, VALUE = 'Single Frame', $
                       UVALUE = 'ANIMNO')
WIDGET_CONTROL, animno, /SET_BUTTON
targanim = 0B

;; Allow user to use data cube times (corresponding to 
;; data records) rather than entered delta-times:
cubebase = WIDGET_BASE(targetbase, /ROW, /EXCLUSIVE)
cubeyes = WIDGET_BUTTON(cubebase, VALUE = 'Use Data Times', $
                        UVALUE = 'CUBEYES')
cubeno = WIDGET_BUTTON(cubebase, VALUE = 'Use Entered Times', $
                       UVALUE = 'CUBENO', SENSITIVE=0)
IF(N_ELEMENTS(data) NE 0) THEN $
                    WIDGET_CONTROL, cubeno, SENSITIVE = 1
cubetimes = 0B

;; Allow user to choose a specific time at which to draw map
mapbase = WIDGET_BASE(targetbase, /ROW, /FRAME)
IF mainuval.cubeloaded THEN $
  CSPICE_ET2UTC, cube[0].et, 'ISOD', 3L, startutc ELSE $
    CSPICE_ET2UTC, mainuval.etstart, 'ISOD', 3L, startutc
maptimewidget = CW_FIELD(mapbase, TITLE = 'Map time', $
                         VALUE = startutc, $
                         UVALUE = 'MAPTIMETEXT', XSIZE = 21, $
                        /RETURN_EVENTS)

;; Allow user to select screen, PS, JPG, or TIFF output and filename:

outputnames = ['Screen', 'PS', 'JPG', 'TIFF']
outputbuttons = CW_BGROUP(targetbase, outputnames, UVALUE = 'OUTPUTBUTTON', $
                    /ROW, /EXCLUSIVE, /RETURN_NAME, SET_VALUE = 0, $
                    UNAME = 'outputbutton', LABEL_LEFT = 'Output type')
screen = 1B
postscript = 0B
jpg = 0B
tiff = 0B

savefilewidget = CW_FIELD(targetbase, TITLE = 'Filename:', UVALUE = $
                        'SAVEFILE', VALUE = 'targetmap', XSIZE = 20)

IF mainuval.targstate.use THEN BEGIN
   savefile = STRTRIM(mainuval.targstate.savefile,2)
   WIDGET_CONTROL, savefilewidget, SET_VALUE = savefile
ENDIF ELSE savefile = ''

IF NOT MAINUVAL.targstate.use THEN BEGIN
    mainuval.targstate.showgrid = grid
    mainuval.targstate.forcecoords = forcecoords
    mainuval.targstate.shadow = shadow
    mainuval.targstate.latlon = latlon
    mainuval.targstate.issnac = issnac
    mainuval.targstate.isswac = isswac
    mainuval.targstate.gridtop = gridtop
    mainuval.targstate.alltimes = alltimes
    mainuval.targstate.animate = targanim
    mainuval.targstate.cubetimes = cubetimes
    mainuval.targstate.screen = screen
    mainuval.targstate.postscript = postscript
    mainuval.targstate.jpg = jpg
    mainuval.targstate.tiff = tiff
    mainuval.targstate.savefile = savefile
ENDIF

IF mainuval.targstate.use THEN BEGIN
    WIDGET_CONTROL, redmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.targstate.spec_max[0]),2)
    WIDGET_CONTROL, greenmaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.targstate.spec_max[1]),2)
    WIDGET_CONTROL, bluemaxwidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.targstate.spec_max[2]),2)
    WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.targstate.ntime),2)
    WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                    STRTRIM(STRING(mainuval.targstate.maptime),2)

    IF mainuval.targstate.animate THEN BEGIN
       WIDGET_CONTROL, animyes, /SET_BUTTON
       targanim = 1B
    ENDIF
    IF mainuval.targstate.alltimes THEN BEGIN
       WIDGET_CONTROL, alltimesbutton, /SET_BUTTON
       alltimes = 1B
    ENDIF
    IF mainuval.targstate.cubetimes THEN BEGIN
       WIDGET_CONTROL, cubeyes, /SET_BUTTON
       cubetimes = 1B
    ENDIF
    IF mainuval.targstate.screen THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
       screen = 1B
    ENDIF
    IF mainuval.targstate.postscript THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 1
       postscript = 1B
       savefile = mainuval.targstate.savefile
    ENDIF
    IF mainuval.targstate.jpg THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 2
       jpg = 1B
       savefile = mainuval.targstate.savefile
    ENDIF
    IF mainuval.targstate.tiff THEN BEGIN
       WIDGET_CONTROL, outputbuttons, SET_VALUE = 3
       tiff = 1B
       savefile = mainuval.targstate.savefile
    ENDIF
    IF NOT mainuval.targstate.latlon THEN BEGIN
       latlon = 0B
       WIDGET_CONTROL, hidelatlonbutton, /SET_BUTTON
    ENDIF
    IF NOT mainuval.targstate.gridtop THEN BEGIN
       gridtop = 0B
       WIDGET_CONTROL, gridonbottom, /SET_BUTTON
    ENDIF
    IF NOT mainuval.targstate.shadow THEN BEGIN
       shadow = 0B
       WIDGET_CONTROL, hideshadowbutton, /SET_BUTTON
    ENDIF
    IF mainuval.targstate.issnac THEN issnac = 1B
    IF mainuval.targstate.isswac THEN isswac = 1B

ENDIF

buttonbase = WIDGET_BASE(targetbase, /ROW)
helpbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Help', $
                           UVALUE = 'HELPBUTTON')
cancelbutton = WIDGET_BUTTON(buttonbase, VALUE = 'Cancel', $
                             UVALUE = 'CANCELBUTTON')
;; Clicking the done button on this widget results in all 
;; values being read from the widget and passed to
;; DIRECT_TARGET.
donebutton = WIDGET_BUTTON(buttonbase, VALUE = 'Make it so', $
                                 UVALUE = 'DONEBUTTON', SENSITIVE=0)

WIDGET_CONTROL, targetbase, /REALIZE, GROUP = geometer_base, /MODAL

IF mainuval.targstate.cubetimes THEN $
       WIDGET_CONTROL, donebutton, SENSITIVE = 1

;; Check this widget to see if the Done button is pressed.
;; Once the done button is pressed, read the values from the
;; various inputs in this widget, destroy the widget, and then
;; send these inputs to the skymap routine.

done_with_targetinput = 0B
spec_bounds = [0,1,2,3,4,5]

REPEAT BEGIN
    targetevent = WIDGET_EVENT()
    CASE targetevent.top OF
        targetbase:BEGIN
            WIDGET_CONTROL, targetevent.id, GET_UVALUE = ev_uval
            CASE ev_uval OF
               'HELPBUTTON': XDISPLAYFILE, mainuval.geometerpath+$
                   'directtargethelp.txt', DONE_BUTTON = 'Done', $
                                           GROUP = targetbase, /MODAL
                'CANCELBUTTON': BEGIN
                    WIDGET_CONTROL, targetevent.top, /DESTROY
                    RETURN
                END
                'DONEBUTTON':BEGIN
                    WIDGET_CONTROL, raminwidget, GET_VALUE = ramin
                    ramin = FLOAT(ramin[0])
                    mainuval.targstate.ramin = ramin
                    WIDGET_CONTROL, ramaxwidget, GET_VALUE = ramax
                    ramax = FLOAT(ramax[0])
                    mainuval.targstate.ramax = ramax
                    WIDGET_CONTROL, deminwidget, GET_VALUE = demin
                    demin = FLOAT(demin[0])
                    mainuval.targstate.decmin = demin
                    WIDGET_CONTROL, demaxwidget, GET_VALUE = demax
                    demax = FLOAT(demax[0])
                    mainuval.targstate.decmax = demax

                    ;; Read out the color selections:
                    WIDGET_CONTROL, redwidget, GET_VALUE = red
                    redlims = STRSPLIT(red,',',/EXTRACT)
                    spec_bounds[0] = FIX(redlims[0])
                    spec_bounds[1] = FIX(redlims[1])
                    WIDGET_CONTROL, redmaxwidget, GET_VALUE = redmaxstring
                    redmax = FLOAT(redmaxstring[0])

                    WIDGET_CONTROL, greenwidget, GET_VALUE = green
                    greenlims = STRSPLIT(green,',',/EXTRACT)
                    spec_bounds[2] = FIX(greenlims[0])
                    spec_bounds[3] = FIX(greenlims[1])
                    WIDGET_CONTROL, greenmaxwidget, GET_VALUE = greenmaxstring
                    greenmax = FLOAT(greenmaxstring[0])

                    WIDGET_CONTROL, bluewidget, GET_VALUE = blue
                    bluelims = STRSPLIT(blue,',',/EXTRACT)
                    spec_bounds[4] = FIX(bluelims[0])
                    spec_bounds[5] = FIX(bluelims[1])
                    WIDGET_CONTROL, bluemaxwidget, GET_VALUE = bluemaxstring
                    bluemax = FLOAT(bluemaxstring[0])

                    mainuval.targstate.spec_bounds = spec_bounds
                    mainuval.targstate.spec_max = [redmax, greenmax, bluemax]

                    WIDGET_CONTROL, ntimewidget, GET_VALUE = ntimestring
                    ntimeplot = FIX(FLOAT(ntimestring[0]))
                    mainuval.targstate.ntime = ntimeplot

                    WIDGET_CONTROL, maptimewidget, GET_VALUE = maptimestring
                    maptime = STRTRIM(maptimestring[0],2)
                    mainuval.targstate.maptime = maptime

                    WIDGET_CONTROL, savefilewidget, GET_VALUE = savefile
                    mainuval.targstate.savefile = savefile

                    done_with_targetinput = 1B
                    mainuval.targstate.use = 1B
                END
                'PLOTBUTTON': BEGIN
                    WIDGET_CONTROL, plotrowselect, GET_VALUE = rowselect
                    rowselect = FIX(FLOAT(rowselect[0])) > 0
                    nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.)) 
                    IF rowselect GE nrows THEN BEGIN
                        WIDGET_CONTROL, plotrowselect, SET_VALUE = $
                                        STRTRIM(STRING(nrows-1),2)
                        rowselect = nrows - 1
                    ENDIF
                    WIDGET_CONTROL, plottimeselect, GET_VALUE = timeselect
                    timeselect = FIX(FLOAT(timeselect[0])) > 0
                    nrecords = N_ELEMENTS(cube.et)
                    IF timeselect GE nrecords THEN BEGIN
                        WIDGET_CONTROL, plottimeselect, SET_VALUE = $
                                        STRTRIM(STRING(nrecords-1), 2)
                        timeselect = nrecords-1
                    ENDIF
                    PLOT, cube[timeselect].data[rowselect, *], XTITLE = $
                          'Channel Number', YTITLE = 'Data Cube Value', $
                          XRANGE = [0, nspec], $
                          TITLE = 'Spectrum for: Record '+$
                          STRTRIM(STRING(timeselect),2)+'. Row '+$
                          STRTRIM(STRING(rowselect),2)+'.', $
                          YRANGE = $
                            [0., MAX(cube[timeselect].data[rowselect,*])]
                 END
                'DATAINFO': BEGIN
                   ;; Create an informational widget displaying summary
                   ;; info on the loaded data cube and on the currently
                   ;; selected row and record. 
                   ;; Get selected row:
                    WIDGET_CONTROL, plotrowselect, GET_VALUE = rowselect
                    rowselect = FIX(FLOAT(rowselect[0])) > 0
                    nrows = N_ELEMENTS(cube[0].data[*,0]) 
                    IF rowselect GE nrows THEN BEGIN
                        WIDGET_CONTROL, plotrowselect, SET_VALUE = $
                                        STRTRIM(STRING(nrows-1),2)
                        rowselect = nrows - 1
                    ENDIF

                    ;; Get selected record:
                    WIDGET_CONTROL, plottimeselect, GET_VALUE = timeselect
                    timeselect = FIX(FLOAT(timeselect[0])) > 0
                    nrecords = N_ELEMENTS(cube.et)
                    IF timeselect GE nrecords THEN BEGIN
                        WIDGET_CONTROL, timeselect, SET_VALUE = $
                                        STRTRIM(STRING(nrecords-1), 2)
                        timeselect = nrecords-1
                    ENDIF

                    ;; Get the selected color ranges:
                    WIDGET_CONTROL, redwidget, GET_VALUE = red
                    redlims = STRSPLIT(red,',',/EXTRACT)
                    spec_bounds[0] = FIX(redlims[0])
                    spec_bounds[1] = FIX(redlims[1])

                    WIDGET_CONTROL, greenwidget, GET_VALUE = green
                    greenlims = STRSPLIT(green,',',/EXTRACT)
                    spec_bounds[2] = FIX(greenlims[0])
                    spec_bounds[3] = FIX(greenlims[1])

                    WIDGET_CONTROL, bluewidget, GET_VALUE = blue
                    bluelims = STRSPLIT(blue,',',/EXTRACT)
                    spec_bounds[4] = FIX(bluelims[0])
                    spec_bounds[5] = FIX(bluelims[1])

                    row_moment = STRING(MOMENT(cube[*].data[$
                                 rowselect,0:nspec-1]))
                    row_max_min = STRTRIM(STRING(MAX(cube[*].data[$
                                  rowselect,0:nspec-1])),2)+' '+STRTRIM(STRING($
                                  MIN(cube[*].data[rowselect,0:nspec-1])),2)

                    record_moment = STRING(MOMENT(cube[timeselect].data[*,*]))
                    rec_max_min = STRTRIM(STRING(MAX(cube[timeselect].data[$
                                  0:nrows-1,0:nspec-1])),2)+' '+$
                                  STRTRIM(STRING($
                                  MIN(cube[timeselect].data[*,*])),2)

                    cube_moment = STRING(MOMENT(cube[*].data[$
                                  0:nrows-1,0:nspec-1]))
                    cube_max_min = STRTRIM(STRING(MAX(cube.data)),2)+' '+$
                                   STRTRIM(STRING(MIN(cube.data[$
                                   0:nrows-1,nspec-1])),2)
                    red_moment = STRING(MOMENT(cube[*].data[*,spec_bounds[0]:$
                                                         spec_bounds[1]]))
                    red_max_min = STRTRIM(STRING(TOTAL(cube[*].data[$
                                  *,spec_bounds[0]:spec_bounds[1]])),2)+$
                                  ' '+STRTRIM(STRING(MIN(cube[*].data[$
                                  *,spec_bounds[0]:spec_bounds[1]])),2)
                    green_moment = STRING(MOMENT(cube[*].data[$
                                   *,spec_bounds[2]:spec_bounds[3]]))
                    green_max_min = STRTRIM(STRING(TOTAL(cube[*].data[$
                                  *,spec_bounds[2]:spec_bounds[3]])),2)+$
                                  ' '+STRTRIM(STRING(MIN(cube[*].data[$
                                  *,spec_bounds[2]:spec_bounds[3]])),2)
                    blue_moment = STRING(MOMENT(cube[*].data[$
                                  *,spec_bounds[4]:spec_bounds[5]]))
                    blue_max_min = STRTRIM(STRING(TOTAL(cube[*].data[$
                                  *,spec_bounds[4]:spec_bounds[5]])),2)+$
                                  ' '+STRTRIM(STRING(MIN(cube[*].data[$
                                  *,spec_bounds[4]:spec_bounds[5]])),2)

                    ;; Put information in a string array for display
                    ;; in the widget.
                    line0 = '                          Mean    '+$
                            '       Std. Dev.        Skewness'+$
                            '       Kurtosis    Maximum  Minimum'
                    line1 = 'Red statistics:   '+red_moment[0]+$
                            red_moment[1]+red_moment[2]+$
                            red_moment[3]+'  '+red_max_min
                    line2 = 'Green statistics: '+green_moment[0]+$
                            green_moment[1]+green_moment[2]+$
                            green_moment[3]+'  '+green_max_min
                    line3 = 'Blue statistics:  '+blue_moment[0]+$
                            blue_moment[1]+blue_moment[2]+$
                            blue_moment[3]+'  '+blue_max_min
                    line4 = 'Row '+STRTRIM(STRING(rowselect),2)+$
                            ' statistics:   '+row_moment[0]+$
                            row_moment[1]+row_moment[2]+$
                            row_moment[3]+'  '+row_max_min
                    line5 = 'Record '+STRTRIM(STRING(timeselect),2)+$
                            ' statistics:'+record_moment[0]+$
                            record_moment[1]+record_moment[2]+$
                            record_moment[3]+'  '+rec_max_min
                    line6 = 'Cube statistics:  '+cube_moment[0]+$
                            cube_moment[1]+cube_moment[2]+$
                            cube_moment[3]+'  '+cube_max_min
                    buffer = [line0, line1, line2, line3, line4, line5, line6]

                    junk = DIALOG_MESSAGE(buffer, TITLE = $
                                          'Data Cube Statistics',$
                                          /CENTER, /INFORMATION)
                END
                'SHADOWYES': BEGIN
                    shadow = 1B
                    mainuval.targstate.shadow = 1B
                END
                'SHADOWNO': BEGIN
                    shadow = 0B
                    mainuval.targstate.shadow = 0B
                END
                'ALLTIMES': BEGIN
                    alltimes = 1B
                    mainuval.targstate.alltimes = 1B
                END
                'ISSNACYES': BEGIN
                    issnac = 1B
                    mainuval.targstate.issnac = issnac
                END
                'ISSNACNO': BEGIN
                  issnac = 0B
                  mainuval.targstate.issnac = issnac
                END
                'ISSWACYES': BEGIN
                  isswac = 1B
                  mainuval.targstate.isswac = isswac
                END
                'ISSWACNO': BEGIN
                  isswac = 1B
                  mainuval.targstate.isswac = isswac
                END
                
                'ONETIME': BEGIN
                    alltimes = 0B
                    mainuval.targstate.alltimes = 0B
                END
                'OUTPUTBUTTON': BEGIN
                   WIDGET_CONTROL, outputbuttons, GET_VALUE = outputtype
                   CASE outputtype OF
                      0: BEGIN
                         screen = 1B
                         postscript = 0B
                         jpg = 0B
                         tiff = 0B
                         mainuval.targstate.screen = 1B
                         mainuval.targstate.postscript = 0B
                         mainuval.targstate.jpg = 0B
                         mainuval.targstate.tiff = 0B
                      END
                      1: BEGIN
                         screen = 0B
                         postscript = 1B
                         jpg = 0B
                         tiff = 0B
                         mainuval.targstate.screen = 0B
                         mainuval.targstate.postscript = 1B
                         mainuval.targstate.jpg = 0B
                         mainuval.targstate.tiff = 0B
                         mainuval.targstate.savefile = savefile
                      END
                      2: BEGIN
                         screen = 0B
                         jpg = 1B
                         postscript = 0B
                         tiff = 0B
                         mainuval.targstate.screen = 0B
                         mainuval.targstate.postscript = 0B
                         mainuval.targstate.jpg = 1B
                         mainuval.targstate.tiff = 0B
                         mainuval.targstate.savefile = savefile
                      END
                      3: BEGIN
                         screen = 0B
                         jpg = 0B
                         postscript = 0B
                         tiff = 1B
                         mainuval.targstate.screen = 0B
                         mainuval.targstate.postscript = 0B
                         mainuval.targstate.jpg = 0B
                         mainuval.targstate.tiff = 1B
                         mainuval.targstate.savefile = savefile
                      END
                   ENDCASE
                END
                'ANIMYES': BEGIN
                   targanim = 1B
                   mainuval.targstate.animate = 1B
                   screen = 1B
                   postscript = 0B
                   jpg = 0B
                   tiff = 0B
                   WIDGET_CONTROL, outputbuttons, SET_VALUE = 0
                END
                'ANIMNO': BEGIN
                   targanim = 0B
                   mainuval.targstate.animate = 0B
                END
                'GRIDYES': BEGIN
                    grid = 1B
                    mainuval.targstate.showgrid = 1B
                END
                'GRIDNO': BEGIN
                    grid = 0B
                    mainuval.targstate.showgrid = 0B
                 END
                'FORCEYES': BEGIN
                   forcecoords = 1B
                   mainuval.targstate.forcecoords = 1B
                END
                'FORCENO': BEGIN
                   forcecoords = 0B
                   mainuval.targstate.forcecoords = 0B
                END
                'GRIDTOP': BEGIN
                    gridtop = 1B
                    mainuval.targstate.gridtop = 1B
                END
                'GRIDBOTTOM': BEGIN
                    gridtop = 0B
                    mainuval.targstate.gridtop = 0B
                END
                'LATLONYES': BEGIN
                    latlon = 1B
                    mainuval.targstate.latlon = 1B
                END
                'LATLONNO': BEGIN
                    latlon = 0B
                    mainuval.targstate.latlon = 0B
                END
                'MAPTIMETEXT': BEGIN
                    WIDGET_CONTROL, maptimewidget, GET_VALUE = $
                      maptimestring
                    CSPICE_UTC2ET, maptimestring[0], maptimeet
                    CSPICE_ET2UTC, maptimeet, 'ISOD', 3L, maptimestring
                    WIDGET_CONTROL, maptimewidget, SET_VALUE = $
                      maptimestring
                END
                'CUBEYES': BEGIN
                    ;; Recalculate plot range boundaries:
                    limbtime = cube[0].et

                    ntimes = N_ELEMENTS(cube.et)
                    IF NOT mainuval.targstate.use THEN $
                       WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                                       STRING(STRTRIM(ntimes),2)

                    ;; Allow user to continue
                    WIDGET_CONTROL, donebutton, SENSITIVE = 1

                    ;; Get the camera rotation matrix used to create 
                    ;; perspective view:
                    CSPICE_SPKPOS, target, limbtime, 'J2000', 'LT', $
                                   'CASSINI', targpos, ltime
                    CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, $
                                   ratarg, dectarg
                    CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, $
                                  3L, 2L, 3L, cam_mat

                    BODY_LIMB, target, 'CASSINI', limbtime, 90, ra_limb, $
                               dec_limb, limberror
                    RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
                             u_limb, v_limb, behind, cam_mat

                    u_min = MIN(u_limb)
                    u_max = MAX(u_limb)
                    u_center = MEAN([u_min,u_max])
                    delta_u = zoomscalefactor*ABS(u_max-u_min)
                    v_min = MIN(v_limb)
                    v_max = MAX(v_limb)
                    v_center = MEAN([v_min, v_max])
                    delta_v = zoomscalefactor*ABS(v_max-v_min) 
                    
                    u_min = u_center+.5*delta_u
                    u_max = u_min-delta_u
                    v_min = v_center-.5*delta_v
                    v_max = v_min+delta_v
                    
                    UV_2_RD, u_min, v_min, ra_min, dec_min, cam_mat
                    UV_2_RD, u_max, v_max, ra_max, dec_max, cam_mat
                    IF NOT mainuval.targstate.use THEN BEGIN
                        WIDGET_CONTROL, raminwidget, SET_VALUE = $
                                        STRTRIM(STRING(ra_min),2)
                        WIDGET_CONTROL, ramaxwidget, SET_VALUE = $
                                        STRTRIM(STRING(ra_max),2)
                        WIDGET_CONTROL, deminwidget, SET_VALUE = $
                                        STRTRIM(STRING(dec_min),2)
                        WIDGET_CONTROL, demaxwidget, SET_VALUE = $
                                        STRTRIM(STRING(dec_max),2)
                    ENDIF
                    cubetimes = 1B
                    mainuval.targstate.cubetimes = 1B
                END
                'CUBENO': BEGIN
                    ;; Recalculate plot range boundaries:
                    limbtime = data[0].et

                    ntimes = N_ELEMENTS(data.ra_full[0])
                    WIDGET_CONTROL, ntimewidget, SET_VALUE = $
                                    STRING(STRTRIM(ntimes),2)

                    ;; Allow user to continue
                    WIDGET_CONTROL, donebutton, SENSITIVE = 1

                    ;; Get the camera rotation matrix used to create 
                    ;; perspective view:
                    CSPICE_SPKPOS, target, limbtime, 'J2000', 'LT', $
                                   'CASSINI', targpos, ltime
                    CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, $
                                   ratarg, dectarg
                    CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, $
                                  3L, 2L, 3L, cam_mat

                    BODY_LIMB, target, 'CASSINI', limbtime, 90, ra_limb, $
                               dec_limb, limberror
                    RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
                             u_limb, v_limb, behind, cam_mat

                    u_min = MIN(u_limb)
                    u_max = MAX(u_limb)
                    u_center = MEAN([u_min,u_max])
                    delta_u = zoomscalefactor*ABS(u_max-u_min)
                    v_min = MIN(v_limb)
                    v_max = MAX(v_limb)
                    v_center = MEAN([v_min, v_max])
                    delta_v = zoomscalefactor*ABS(v_max-v_min) 
                    
                    u_min = u_center+.5*delta_u
                    u_max = u_min-delta_u
                    v_min = v_center-.5*delta_v
                    v_max = v_min+delta_v
                    
                    UV_2_RD, u_min, v_min, ra_min, dec_min, cam_mat
                    UV_2_RD, u_max, v_max, ra_max, dec_max, cam_mat
                    IF NOT mainuval.targstate.use THEN BEGIN
                        WIDGET_CONTROL, raminwidget, SET_VALUE = $
                                        STRTRIM(STRING(ra_min),2)
                        WIDGET_CONTROL, ramaxwidget, SET_VALUE = $
                                        STRTRIM(STRING(ra_max),2)
                        WIDGET_CONTROL, deminwidget, SET_VALUE = $
                                        STRTRIM(STRING(dec_min),2)
                        WIDGET_CONTROL, demaxwidget, SET_VALUE = $
                                        STRTRIM(STRING(dec_max),2)
                    ENDIF
                    cubetimes = 0B
                    mainuval.targstate.cubetimes = 0B
                END
                ELSE:
            ENDCASE
        END
        ELSE: ;; Do nothing
    ENDCASE
ENDREP UNTIL done_with_targetinput
WIDGET_CONTROL, targetbase, /DESTROY

IF cubetimes THEN ntimes = N_ELEMENTS(cube.et) ELSE $
   ntimes = N_ELEMENTS(data.ra_full[0])

IF targanim THEN XINTERANIMATE, SET = [600, 600, ntimeplot] 

;; Identify the time to make the map from the user entry:
IF NOT cubetimes THEN BEGIN
;    WIDGET_CONTROL, mainuval.maptimetext, GET_VALUE = maptime
;    maptime = maptime[0]
    CSPICE_UTC2ET, maptime, mapet
    junk = MIN(ABS(mapet-data.et),etloc)
    CSPICE_ET2UTC, data[etloc].et, mainuval.tformat, 3L, maptime
    mainuval.targstate.maptime = maptime
    WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                    'Actual time for map: ' + maptime
ENDIF ELSE BEGIN
   CSPICE_UTC2ET, maptime, mapet
   junk = MIN(ABS(mapet-cube.et), etloc)
   CSPICE_ET2UTC, cube[etloc].et, mainuval.tformat, 3L, maptime
   mainuval.targstate.maptime = maptime
ENDELSE

;; Get star limiting magnitude:
WIDGET_CONTROL, mainuval.mapstarlimtext, GET_VALUE = starlim
starlim = FLOAT(starlim[0])

;; Determine which set of times to use and store in vectors
;; to pass to DIRECT_TARGET.

dtime_ix = FLOAT(ntimes)/ntimeplot
plot_ets = DBLARR(ntimeplot)
plot_utcs = STRARR(ntimeplot)

IF cubetimes THEN BEGIN
    IF mainuval.cubeloaded THEN BEGIN
        cube_target = cube
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = cube[gettime_ix].et
            CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                           3L, utc0
            plot_utcs[time_ix] = utc0
        ENDFOR
    ENDIF ELSE BEGIN
        FOR time_ix = 0, ntimeplot-1 DO BEGIN
            gettime_ix = time_ix*dtime_ix < (ntimes-1)
            plot_ets[time_ix] = data[gettime_ix].et
            plot_utcs[time_ix] = data[gettime_ix].utc
        ENDFOR
        WIDGET_CONTROL, mainuval.statustext, SET_VALUE = $
                        'Cube not loaded.'
        WIDGET_CONTROL, targetbase, /DESTROY
        RETURN
    ENDELSE
ENDIF ELSE BEGIN
    ;; Create a dummy variable to take the place of the cube
    ;; data to pass to DIRECT_TARGET since cube data may
    ;; not be loaded if we're here.
    cube_target = {et:FLTARR(1)+1E20} 
    ;; Give it a ridiculous value so
    ;; DIRECT_TARGET will know it's not real.

    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        gettime_ix = time_ix*dtime_ix < (ntimes-1)
        plot_ets[time_ix] = data[gettime_ix].et
        CSPICE_ET2UTC, plot_ets[time_ix], mainuval.tformat, $
                       3L, utc0
        plot_utcs[time_ix] = utc0
    ENDFOR
ENDELSE

;; Now make the call(s) to DIRECT_TARGET
IF targanim THEN BEGIN
    maptimedummy = ''
    FOR time_ix = 0, ntimeplot-1 DO BEGIN
        data_ix = time_ix*dtime_ix < (ntimes-1)
        DIRECT_TARGET, data, mainuval, data_ix, time_ix, plot_ets, $
                       plot_utcs, cube_target, spec_bounds, maptimedummy, $
                       RA_RANGE = [ramin, ramax], DEC_RANGE = $
                       [demin, demax], COLORMAX = [redmax, greenmax, bluemax], $
                       ALLTIMES = alltimes, STARS = starlim, /ANIMATE, $
                       GRID = grid, SHADOW = shadow, $
                       LATLON = latlon, TOPGRID = gridtop, $
                       SCREEN = screen, POSTSCRIPT = postscript, $
                       JPG = jpg, TIFF = tiff, SAVEFILE = savefile, $
                       FORCECOORDS = forcecoords, NAC = issnac, WAC = isswac

        XINTERANIMATE, FRAME = time_ix, WINDOW = !D.WINDOW
    ENDFOR
    XINTERANIMATE
ENDIF ELSE $
  DIRECT_TARGET, data, mainuval, etloc, etloc, plot_ets, $
                 plot_utcs, cube_target, spec_bounds, maptime, $
                 RA_RANGE = [ramin, ramax], DEC_RANGE = [demin, demax], $
                 ALLTIMES = alltimes, STARS = starlim, COLORMAX = $
                 [redmax, greenmax, bluemax], GRID = grid, SHADOW = shadow, $
                 LATLON = latlon, TOPGRID = gridtop, $
                 SCREEN = screen, POSTSCRIPT = postscript, $
                 JPG = jpg, TIFF = tiff, SAVEFILE = savefile, $
                 FORCECOORDS = forcecoords, NAC = issnac, WAC = isswac


END
