;;File Version: 1.4
;;Last Modified: June 15 , 2018
;;By: JEC

PRO DIRECT_TARGET, data, maininfo, i_time, i_plot, plot_ets, plot_utcs, $
                   cube, spec_bounds, maptime, STARS = STARS, $
                   ALLTIMES = ALLTIMES, RA_RANGE = RA_RANGE, $
                   DEC_RANGE = DEC_RANGE, ANIMATE = ANIMATE, $
                   COLORMAX = COLORMAX, GRID = GRID, SHADOW = SHADOW, $
                   LATLON = LATLON, TOPGRID = TOPGRID, $
                   SCREEN = SCREEN, POSTSCRIPT = POSTSCRIPT, $
                   JPG = JPG, TIFF = TIFF, SAVEFILE = SAVEFILE, $
                   FORCECOORDS = FORCECOORDS, NAC = NAC, WAC = WAC

COMPILE_OPT idl2
;;  >> Abstract <<
;; Only to be called by CALL_TARGET.
;; Draws objects in the "U, V" coordinate system, which is the
;; sky plane coordinate system in normalized coordinates. The
;; sky plane is defined as being the plane normal to the line of sight
;; from Cassini to the center of the target body.
;; i_time is the index of the desired time for computing body data.
;; maininfo is the UVALUE structure from GEOMETER.
;; Keywords:
;; STARS: set equal to the limiting magnitude of stars to be drawn.
;; If it is set to a magnitude of less than -1 then stars will not
;; be drawn.
;; ALLTIMES: If set, pixels are drawn for all times for which data exist.
;; ANIMATE: If set, use the first time value for determining the plot
;; range and grid values to avoid a bouncing coordinate grid in the 
;; animation.
;; TOPGRID: redraw the grid after the pixels have been drawn.
;; Author:
;; Josh Colwell
;; Based on GEOMETER_TARGET.
;;
;; January 17, 2007: added maptime option. AWD.
;; February 8, 2007: Modify for version 2 of Data Cube.
;; March 1, 2007: Modify for Geometer version of data cube.
;; May 22, 2007: Fix platform-dependent device bug for set_plot.
;; May 25, 2007: Handle F Ring target. Display Target distance on plot.
;; September 2, 2007: Add Titan exosphere.
;; December 18, 2007: Updated to allow multiple output file formats. AWD. 
;; March 12, 2008: Minor bug fixes in updated output. JEC.
;; January 24, 2010: Add the FORCECOORDS keyword. Make it so that when
;; plotting data cubes, only data for those pixels selected in the
;; main pixel selection window are plotted. JEC.
;; Fix bug when terminator data are not more than one point in the vectors. JEC.
;; Some color plotting and other updates to fix some cosmetic bugs and update font size. JEC.
;; June 2018: Several cosmetic changes and bug fixes. Previously was not
;; drawing shadow stipple on anything other than Saturn. This seems to be fixed. Some
;; fixes to colors. Fixed some issues with writing to JPEG and TIFF. JEC.

;;Identify the time to make the map from the user entry
IF maptime NE '' THEN CSPICE_UTC2ET, maptime, mapet

IF KEYWORD_SET(NAC) THEN NAC = 1B ELSE NAC = 0B
IF KEYWORD_SET(WAC) THEN WAC = 1B ELSE WAC = 0B

IF NOT KEYWORD_SET(POSTSCRIPT) THEN BEGIN
   red = 255                    ; Define some colors for later
   orange = 107000                
   yellow = 6553599 
   green = 65280
   cyan = 16711500
   blue = 16711680
   violet = 16600000
   white = 256L*256L*256L-1
   black = 0

   !P.THICK = 1
   !P.CHARTHICK = 1

;; Make the plot window:

   SET_PLOT, 'X'

   IF KEYWORD_SET(SCREEN) THEN winsize = 600

   IF (KEYWORD_SET(JPG) OR KEYWORD_SET(TIFF)) THEN BEGIN
      !P.BACKGROUND = WHITE
      !P.COLOR = BLACK
      winsize = 1000
   ENDIF

   WINDOW, 0, XSIZE = winsize, YSIZE = winsize

ENDIF ELSE BEGIN
; Color table written by Liam.Gumley@ssec.wisc.edu
; http://cow.physics.wisc.edu/~craigm/idl/archive/msg00492.html
   r = [0,255,0,255,0,255,0,255,0,255,255,112,219,127,0,255]
   g = [0,0,255,255,255,0,0,255,0,187,127,219,112,127,163,171]
   b = [0,255,255,0,0,0,255,255,115,0,127,147,219,127,255,127]
   TVLCT, r, g, b
   red = 5
   orange = 9
   yellow = 3
   green = 4
   cyan = 2
   blue = 6
   violet = 12
   white = 7

   SET_PLOT, 'PS'

   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF (extension NE '.ps') THEN savefile = savefile + '.ps'

   DEVICE, FILENAME=savefile, XSIZE=20, YSIZE=20, /ENCAPSULATED, $
           /COLOR, BITS_PER_PIXEL=8
   !P.THICK = 2
   !P.CHARTHICK = 2

ENDELSE

labelfontsize = 1.3

; Make a circle symbol for plotting
circ = findgen(17) * (!PI*2/16.)
usersym, COS(circ), SIN(circ)

;;Get the string value for the target body:
targetid = WIDGET_INFO(maininfo.targetdrop, /DROPLIST_SELECT)
WIDGET_CONTROL, maininfo.targetdrop, GET_VALUE = targetdropvalue
target = STRCOMPRESS(STRUPCASE(targetdropvalue[targetid]),/REMOVE_ALL)
CSPICE_BODN2C, target, targetid, found
IF NOT found THEN BEGIN
    target = 'SATURN'
    targetid = 699L 
    ;;This should only happen if the target is the rings.
ENDIF

IF cube[0].et GT 1E19 THEN BEGIN
   ;; cube.et is set to 1E20 if cube data is not present or not to be used.
   npixel = N_ELEMENTS(data[i_time].ra_full[0,*])
   selected_pixels = WHERE(maininfo.pxvec)
   name = 'Row '+STRTRIM(STRING(selected_pixels[0]),2)
   channel_name = 'UVIS'
   IF maptime EQ '' THEN $
      time = data[i_time].et $
   ELSE time = mapet
   
   IF KEYWORD_SET(ANIMATE) THEN gridtime = data[0].et ELSE gridtime = time
ENDIF ELSE BEGIN
   npixel = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
   name = 'UVIS'
   channel_name = 'UVIS'
    IF maptime EQ '' THEN $
       time = plot_ets[i_plot] $
    ELSE time = mapet
   IF KEYWORD_SET(ANIMATE) THEN gridtime = plot_ets[0] ELSE gridtime = time
ENDELSE

;; Get the camera rotation matrix used to create perspective view:
CSPICE_SPKPOS, target, time, 'J2000', 'LT', 'CASSINI', targpos, ltime
CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, ratarg, dectarg
CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, 3L, 2L, 3L, cam_mat

;; For the initial plot size, use a cam_mat that corresponds to the initial
;; time if animating:
CSPICE_SPKPOS, target, gridtime, 'J2000', 'LT', 'CASSINI', targpos2, ltime2
CSPICE_RECRAD, targpos2/CSPICE_VNORM(targpos2), unit, ratarg2, dectarg2
CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg2, ratarg2, 3L, 2L, 3L, cam_mat_grid

IF name EQ 'Row 64' THEN BEGIN
    CASE maininfo.channel OF
        1: channel_name = 'UVIS EUV'
        2: channel_name = 'UVIS EUV'
        3: channel_name = 'UVIS Solar'
        4: channel_name = 'UVIS FUV'
        5: channel_name = 'UVIS FUV'
        6: channel_name = 'UVIS FUV'
        7: channel_name = 'UVIS HSP'
        8: channel_name = 'UVIS HDAC'
        9: channel_name = 'CIRS FP1'
        10: channel_name = 'CIRS FP3'
        11: channel_name = 'CIRS FP4'
        12: channel_name = 'ISS NAC'
        13: channel_name = 'ISS WAC'
        ELSE: channel_name = 'UVIS'
    ENDCASE
ENDIF

ra_min = RA_RANGE[0]
ra_max = RA_RANGE[1]
dec_min = DEC_RANGE[0]
dec_max = DEC_RANGE[1]
RD_2_UV, ra_min, dec_min, u_min, v_min, behind, cam_mat_grid
RD_2_UV, ra_max, dec_max, u_max, v_max, behind, cam_mat_grid

IF NOT KEYWORD_SET(FORCECOORDS) THEN BEGIN
   delta_v = ABS(u_max-u_min)
   v_min = -delta_v/2.
   v_max = delta_v/2.
ENDIF

PLOT, [[0,0],[0,0]], /NODATA, XRANGE = [u_min, u_max], YRANGE = $
      [v_min, v_max], XSTYLE = 4+1, YSTYLE = 4+1

;;*************************************************************************
;;                              BODIES
;;*************************************************************************
;; First draw the target body:
BODY_LIMB, target, 'CASSINI', time, 90, ra_limb, dec_limb, limberror

IF NOT limberror THEN BEGIN
    RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
             u_limb, v_limb, behind, cam_mat
    IF NOT behind THEN OPLOT, u_limb, v_limb
ENDIF ELSE BEGIN
    WIDGET_CONTROL, maininfo.statustext, SET_VALUE = 'Error in DIRECT_TARGET'
    RETURN
ENDELSE
;XYOUTS, MEAN(u_limb), MEAN(v_limb), target, SIZE = labelfontsize

;; Do the Titan exobase:
TITAN_EXO_LIMB, 'CASSINI', time, 36, ra_exo, dec_exo
RD_2_UV, [ra_exo, ra_exo[0]], [dec_exo, dec_exo[0]], $
          u_exo, v_exo, behind, cam_mat
IF NOT behind THEN OPLOT, u_exo, v_exo

;; *******************************TERMINATOR********************************
;; Draw the terminator on the target body:
ra_term = [0D0]
dec_term = [0D0]
CSPICE_BODVAR, targetid, 'RADII', radii

CSPICE_SPKPOS, target, time, 'J2000', 'LT', 'CASSINI', scpos, ltime
CSPICE_SPKPOS, target, time, 'J2000', 'LT', 'SUN', sunpos, ltimesun
CSPICE_EDLIMB, radii[0], radii[1], radii[2], sunpos, term
;; term is a SPICE ellipse defining the terminator.
CSPICE_EL2CGV, term, center, v1, v2

;; Get RA/Dec points along the terminator of the body:
FOR j = 0, 71 DO BEGIN
    AZ = !DPI*2.0/71D0*j-!DPI
    VEC2 = CENTER+COS(AZ)*V1+SIN(AZ)*V2
    termvec = scpos-vec2
    ;; This is actually a less then fully accurate check to
    ;; see if the terminator line is behind the limb...
    IF CSPICE_VNORM(termvec) LT CSPICE_VNORM(scpos) THEN BEGIN
        CSPICE_RECRAD, termVEC, RANGE, RA1, DEC1
        ra_term = [ra_term, RA1*!RADEG]
        dec_term = [dec_term, DEC1*!RADEG]
    ENDIF
ENDFOR
IF N_ELEMENTS(ra_term) GT 1 THEN BEGIN
  ra_term = ra_term[1:*]
  dec_term = dec_term[1:*]
  ;; Remove discontinuity across the body:
  delta_ra_term = ra_term-SHIFT(ra_term,1)
  delta_ra_term = delta_ra_term[1:*]
  void = MAX(ABS(delta_ra_term),breakloc)
  ;; Rearrange to avoid drawing across the body:
  targ_ra_size = MAX(ra_limb)-MIN(ra_limb)
  IF (void GT targ_ra_size/2.) AND (breakloc NE 0) THEN BEGIN
    ra_term_temp = [ra_term[breakloc:*], ra_term[0:breakloc-1]]
    dec_term_temp = [dec_term[breakloc:*], dec_term[0:breakloc-1]]
    ra_term = ra_term_temp[1:*]
    dec_term = dec_term_temp[1:*]
  ENDIF

  delta_dec_term = dec_term-SHIFT(dec_term,1)
  delta_dec_term = delta_dec_term[1:*]
  void = MAX(ABS(delta_dec_term),breakloc)
  ;; Rearrange to avoid drawing across the body:
  targ_dec_size = MAX(dec_limb)-MIN(dec_limb)
  IF (void GT targ_dec_size/2.) AND (breakloc NE 0) THEN BEGIN
    ra_term_temp = [ra_term[breakloc:*], ra_term[0:breakloc-1]]
    dec_term_temp = [dec_term[breakloc:*], dec_term[0:breakloc-1]]
    ra_term = ra_term_temp[1:*]
    dec_term = dec_term_temp[1:*]
  ENDIF
  RD_2_UV, ra_term, dec_term, u_term, v_term, behind, cam_mat
  IF NOT behind THEN OPLOT, u_term, v_term, COLOR = red
ENDIF

;; ************************LATITUDE/LONGITUDE*****************************
CSPICE_PXFORM, 'J2000', 'IAU_'+target, time, inert2p
IF KEYWORD_SET(LATLON) THEN BEGIN
;; To get the Lat/Lon points: get vectors from body center to lat/lon
;; positions, then add those to vector from S/C to planet, convert to RA/Dec
;; and then convert to U/V to plot.
ndrawpoints = 100
u_lat_vec = DBLARR(ndrawpoints)
v_lat_vec = DBLARR(ndrawpoints)
visible_vec = BYTARR(ndrawpoints)
long_int = 30D0*!DTOR
lat_int = 15D0*!DTOR
lat_min = -!DPI/2.+lat_int
lat_max = !DPI/2-lat_int
nlatlines = (lat_max-lat_min)/lat_int+1

;; need these stored for the topgrid option later:
scposlatlon = scpos
inert2platlon = inert2p
cam_mat_latlon = cam_mat
FOR lat_ix = 0, nlatlines-1 DO BEGIN
    latvec = DBLARR(ndrawpoints)+lat_min+lat_ix*lat_int
    lonvec = DINDGEN(ndrawpoints)/(ndrawpoints-1D0)*!DPI*2D0
    CSPICE_LATREC, DBLARR(ndrawpoints)+radii[0], lonvec, latvec, $
                   latlonpos

    ;; I think I can get rid of the following loop by using vectorized
    ;; CSPICE calls, but just get it working for now and worry about
    ;; performance later.
    FOR vec_ix = 0, ndrawpoints-1 DO BEGIN
        latlonvec = REFORM(latlonpos[*,vec_ix])
        latlonvec /= CSPICE_VNORM(latlonvec)
        latlonvec *= SQRT((radii[0]*COS(latvec[lat_ix]))^2+$
                          (radii[2]*SIN(latvec[lat_ix]))^2)
        ;; latlonvec is the vector to the particular point at this 
        ;; latitude and longitude on the surface of oblate spheroid
        ;; target (note, not coded for triaxial ellipsoid, just oblate
        ;; spheroid).
        CSPICE_MTXV, inert2platlon, latlonvec, latlonvec_inert
        vec2point = latlonvec_inert-scposlatlon
        IF CSPICE_VNORM(vec2point) LE CSPICE_VNORM(scposlatlon) THEN $
          visible_vec[vec_ix] = 1 ELSE visible_vec[vec_ix] = 0
        CSPICE_RECRAD, vec2point, dist, ra_pnt, dec_pnt
        RD_2_UV, ra_pnt*!RADEG, dec_pnt*!RADEG, u_pnt, v_pnt, behind, $
                 cam_mat_latlon
        u_lat_vec[vec_ix] = u_pnt
        v_lat_vec[vec_ix] = v_pnt
    ENDFOR
    good = WHERE(visible_vec)
    IF N_ELEMENTS(good) GT 2 THEN BEGIN
        delta_good = good-SHIFT(good,1)
        junk = MAX(delta_good, breakloc)
        good_us = u_lat_vec[good[0:breakloc-1]]
        good_vs = v_lat_vec[good[0:breakloc-1]]
        latlabel_u = MEAN(good_us)
        junk = MIN(ABS(latlabel_u-good_us), loc)
        latlabel_v = good_vs[loc]
        IF lat_ix EQ FIX(nlatlines/2) THEN $
          OPLOT, good_us, good_vs, COLOR = white ELSE $
            OPLOT, good_us, good_vs, COLOR = cyan
        good_us = u_lat_vec[good[breakloc:*]]
        good_vs = v_lat_vec[good[breakloc:*]]
        
        ;; Draw the equator in white and the other latitude lines in cyan:
        IF lat_ix EQ FIX(nlatlines/2) THEN $
          OPLOT, good_us, good_vs, COLOR = white ELSE $
            OPLOT, good_us, good_vs, COLOR = cyan

        ;; Only label some latitude lines:
        CASE lat_ix OF
          FIX(nlatlines/2-1): XYOUTS, latlabel_u, latlabel_v, $
            STRMID(STRING(STRTRIM(ROUND(lat_int*!RADEG),2)),0,2), SIZE = labelfontsize
          FIX(nlatlines/2-1)-2: XYOUTS, latlabel_u, latlabel_v, $
            STRMID(STRING(STRTRIM(ROUND(lat_int*3*!RADEG),2)),0,2), SIZE = labelfontsize
          FIX(nlatlines/2+1): XYOUTS, latlabel_u, latlabel_v, '-'+$
            STRMID(STRING(STRTRIM(ROUND(lat_int*!RADEG),2)),0,2), SIZE = labelfontsize
          FIX(nlatlines/2+1)+2: XYOUTS, latlabel_u, latlabel_v, '-'+$
            STRMID(STRING(STRTRIM(ROUND(lat_int*3*!RADEG),2)),0,2), SIZE = labelfontsize
          ELSE: ;; don't label this latitude line.
        ENDCASE
    ENDIF
ENDFOR

;; Same thing for longitude:
u_lon_vec = DBLARR(ndrawpoints)
v_lon_vec = DBLARR(ndrawpoints)
lon_min = 0D0
lon_max = !DPI*2.-long_int
nlonlines = (lon_max-lon_min)/long_int+1
FOR lon_ix = 0, nlonlines-1 DO BEGIN
    lonvec = DBLARR(ndrawpoints)+lon_min+lon_ix*long_int
    latvec = DINDGEN(ndrawpoints)/(ndrawpoints-1D0)*!DPI-!DPI/2D0
    CSPICE_LATREC, DBLARR(ndrawpoints)+radii[0], lonvec, latvec, $
                   latlonpos

    ;; I think I can get rid of the following loop by using vectorized
    ;; CSPICE calls, but just get it working for now and worry about
    ;; performance later.
    FOR vec_ix = 0, ndrawpoints-1 DO BEGIN
        latlonvec = REFORM(latlonpos[*,vec_ix])
        latlonvec /= CSPICE_VNORM(latlonvec)
        latlonvec *= SQRT((radii[0]*COS(latvec[lat_ix]))^2+$
                          (radii[2]*SIN(latvec[lat_ix]))^2)
        ;; latlonvec is the vector to the particular point at this 
        ;; latitude and longitude on the surface of oblate spheroid
        ;; target (note, not coded for triaxial ellipsoid, just oblate
        ;; spheroid).
        CSPICE_MTXV, inert2platlon, latlonvec, latlonvec_inert
        vec2point = latlonvec_inert-scposlatlon
        IF CSPICE_VNORM(vec2point) LE CSPICE_VNORM(scposlatlon) THEN $
          visible_vec[vec_ix] = 1 ELSE visible_vec[vec_ix] = 0
        CSPICE_RECRAD, vec2point, dist, ra_pnt, dec_pnt
        RD_2_UV, ra_pnt*!RADEG, dec_pnt*!RADEG, u_pnt, v_pnt, behind, $
                 cam_mat_latlon
        u_lon_vec[vec_ix] = u_pnt
        v_lon_vec[vec_ix] = v_pnt
    ENDFOR
    good = WHERE(visible_vec)
    IF N_ELEMENTS(good) GT 2 THEN BEGIN
        delta_good = good-SHIFT(good,1)
        junk = MAX(delta_good, breakloc)
        good_us = u_lon_vec[good[0:breakloc-1]]
        good_vs = v_lon_vec[good[0:breakloc-1]]
        OPLOT, good_us, good_vs, COLOR = cyan
        good_us = u_lon_vec[good[breakloc:*]]
        good_vs = v_lon_vec[good[breakloc:*]]
        OPLOT, good_us, good_vs, COLOR = cyan
        lon_v_label = MEAN(good_vs)
        junk = MIN(ABS(lon_v_label-good_vs), loc)
;; Experience shows labeling longitude lines can lead to very cluttered plots
;; so that is commented out.
;        IF lon_ix EQ 0 THEN $
;          XYOUTS, good_us[loc], lon_v_label, '0' ELSE $
;            IF lon_ix EQ FIX(nlonlines/2-1) THEN $
;              XYOUTS, good_us[loc], lon_v_label, '180'
    ENDIF
ENDFOR
ENDIF  ;; LATLON keyword set.

;; **************************INTERCEPT POINTS*******************************
;; Plot the sub-solar and sub-spacecraft points, and the poles:
CSPICE_RECRAD, scpos, dist, ra, dec
RD_2_UV, ra*!RADEG, dec*!RADEG, usub, vsub, behind, cam_mat
PLOTS, [usub], [vsub], PSYM = 4, COLOR = green

CSPICE_SUBSOL, 'NEAR POINT', target, time, 'LT', 'CASSINI', subsolpoint
CSPICE_MTXV, inert2p, subsolpoint, subsol_inert
vec2subsol = scpos+subsol_inert

;; Only plot it if it's on the near side of the planet:
IF CSPICE_VNORM(vec2subsol) LE CSPICE_VNORM(scpos) THEN BEGIN 
    CSPICE_RECRAD, vec2subsol, dist, ra, dec
    RD_2_UV, ra*!RADEG, dec*!RADEG, usub, vsub, behind, cam_mat
    PLOTS, [usub], [vsub], PSYM = 4, COLOR = red
    XYOUTS, usub, vsub, 'Sub-Solar', SIZE = labelfontsize
ENDIF

target_pole_north_body = [0D0, 0D0, 1D0]
target_pole_south_body = [0D0, 0D0, -1D0]
CSPICE_MTXV, inert2p, target_pole_north_body, target_north
CSPICE_MTXV, inert2p, target_pole_south_body, target_south
vec2northpole = scpos+target_north*radii[2]
vec2southpole = scpos+target_south*radii[2]
IF CSPICE_VNORM(vec2northpole) LE CSPICE_VNORM(scpos) THEN BEGIN
    CSPICE_RECRAD, vec2northpole, dist, north_ra, north_dec
    RD_2_UV, north_ra*!RADEG, north_dec*!RADEG, unorth, vnorth, behind, cam_mat
    PLOTS, [unorth], [vnorth], PSYM=2, COLOR = blue
    XYOUTS, unorth, vnorth, 'North', SIZE = labelfontsize
ENDIF
;IF CSPICE_VNORM(vec2southpole) LE CSPICE_VNORM(scpos) THEN BEGIN
;    CSPICE_RECRAD, vec2southpole, dist, south_ra, south_dec
;    RD_2_UV, south_ra*!RADEG, south_dec*!RADEG, usouth, vsouth, behind, cam_mat
;    PLOTS, [usouth], [vsouth], PSYM=2, COLOR = blue
;    XYOUTS, usouth, vsouth, 'South', SIZE = labelfontsize
;ENDIF



;;;---------------------------------
;; ***************************SHADOW ON TARGET*****************************
IF KEYWORD_SET(SHADOW) THEN BEGIN
  u1 = MIN(u_limb)
  u2 = MAX(u_limb)
  v1 = MIN(v_limb)
  v2 = MAX(v_limb)
  urange = u2-u1
  vrange = v2-v1
  nshadpoints = 40
  delta_u = urange/nshadpoints
  delta_v = vrange/nshadpoints
  subsol_from_inside = vec2subsol-scpos
  FOR u_ix = 0, nshadpoints-1 DO BEGIN
    u_pnt = u1+u_ix*delta_u
    FOR v_ix = 0, nshadpoints-1 DO BEGIN
      v_pnt = v1+v_ix*delta_v
      ;; Check to see if point is on the body and in the shadow.
      UV_2_RD, u_pnt, v_pnt, ra_pnt, dec_pnt, cam_mat
      CSPICE_RADREC, 1D0, ra_pnt*!DTOR, dec_pnt*!DTOR, rec_pnt
      CSPICE_SURFPT, -scpos, rec_pnt, radii[0], radii[1], radii[2], $
        point, found
      IF found THEN BEGIN
        ;; It's on the body, now check to see if it's on the unlit
        ;; side.
        lit = CSPICE_VSEP(point, subsol_from_inside) LE !DPI/2
        IF NOT lit THEN PLOTS, u_pnt, v_pnt, COLOR = red, PSYM=3
      ENDIF
    ENDFOR
  ENDFOR
ENDIF

;;*************************************************************************
;;                              RINGS
;;*************************************************************************
;; Only do this if Saturn is not behind the viewer:
CSPICE_SPKPOS, 'SATURN', time, 'J2000', 'LT', 'CASSINI', $
               scpos, lt_corr
CSPICE_RECRAD, scpos, dist, rasat, decsat
RD_2_UV, rasat*!RADEG, decsat*!RADEG, usat, vsat, satbehind, cam_mat
IF NOT satbehind THEN BEGIN 
;; Do Saturn's rings. For now, hardwire the critical ring boundaries.
;; Would be better eventually to read these values in from PCK.
ringrad = [66970.,74510.,92000.,117580.,122170.,133405.,133730.,136780.,$
           140180.,140270.]
nring = N_ELEMENTS(ringrad)    
    
;; Get the rotation matrix between J2000 and Saturn:
CSPICE_PXFORM, 'J2000', 'IAU_SATURN', time-lt_corr, inert2p
inert2p = TRANSPOSE(inert2p)
scpos_p = inert2p#scpos
CSPICE_BODVAR, 699L, 'RADII', satrad

nringpoints = 360
;;Use 360 points for each ring

;; Store the extrema of u and v values for the rings for shadow drawing:
uringmin = 1E9
uringmax = -1E9
vringmin = 1E9
vringmax = -1E9

FOR ringindex = 0, nring-1 DO BEGIN
  ringra = DBLARR(nringpoints)
  ringdec = DBLARR(nringpoints)

  FOR ringpoint = 0, nringpoints-1 DO BEGIN
      az = !DPI*2.0*DOUBLE(ringpoint)/DOUBLE(nringpoints)-!DPI
      rvec = [COS(az), SIN(az), 0.0]*ringrad[ringindex]
      ;; RVEC is the vector to the ring in body coordinates.
      
      vec = TRANSPOSE(inert2p)#(scpos_p+rvec)
      ;; VEC is the vector from the S/C to the ring in IREF.
      ;; Check to see if it intercepts the planet.
      
      CSPICE_SURFPT,-scpos_p,inert2p#VEC,satrad[0],satrad[1],satrad[2], $
                    POINT, found
      HIDDEN = 0B
      IF found THEN BEGIN
          ;; See if the point is behind the planet.
          IF (CSPICE_VNORM(SCPOS) LT CSPICE_VNORM(VEC)) THEN HIDDEN = 1B
      ENDIF
      IF (NOT HIDDEN) THEN BEGIN
          CSPICE_RECRAD, VEC, RANGE, RA1, DEC1
          ringra[ringpoint] = RA1*!RADEG
          ringdec[ringpoint] = DEC1*!RADEG
      ENDIF
  ENDFOR    ;;ringpoint

  ;; Hidden parts of the rings have coordinate values of [0.,0.].
  ;; Identify which points are not hidden:
  good = WHERE(ringra NE 0D0)
  ;; Find where there is a break in good:
  void = MAX(good-SHIFT(good,1), breakloc)
  ;; Rearrange the order of points so the ring is drawn without 
  ;; connecting behind the body:
  IF void GT 1 THEN good = [good[breakloc:*],good[0:breakloc-1]] ELSE $
    IF N_ELEMENTS(good) EQ nringpoints THEN good = [good, good[0]]
  
  RD_2_UV, ringra[good], ringdec[good], ringu, ringv, behind, cam_mat
  OPLOT, ringu, ringv, COLOR = orange

  ;; Update the ring UV extrema:
  uringmin = uringmin < MIN(ringu)
  uringmax = uringmax > MAX(ringu)
  vringmin = vringmin < MIN(ringv)
  vringmax = vringmax > MAX(ringv)
ENDFOR  ;;ringindex
ENDIF   ;;Saturn and rings not behind viewer.

;; ***************************SHADOW ON TARGET*****************************
IF KEYWORD_SET(SHADOW) THEN BEGIN
    u1 = MIN(u_limb)
    u2 = MAX(u_limb)
    v1 = MIN(v_limb)
    v2 = MAX(v_limb)
    urange = u2-u1
    vrange = v2-v1
    nshadpoints = 40
    delta_u = urange/nshadpoints
    delta_v = vrange/nshadpoints
    subsol_from_inside = vec2subsol-scpos
    FOR u_ix = 0, nshadpoints-1 DO BEGIN
        u_pnt = u1+u_ix*delta_u
        FOR v_ix = 0, nshadpoints-1 DO BEGIN
            v_pnt = v1+v_ix*delta_v
            ;; Check to see if point is on the body and in the shadow.
            UV_2_RD, u_pnt, v_pnt, ra_pnt, dec_pnt, cam_mat
            CSPICE_RADREC, 1D0, ra_pnt*!DTOR, dec_pnt*!DTOR, rec_pnt
            CSPICE_SURFPT, -scpos, rec_pnt, radii[0], radii[1], radii[2], $
                           point, found
;IF u_ix eq 20 THEN stop,'is it on body?'
            IF found THEN BEGIN
                ;; It's on the body, now check to see if it's on the unlit
                ;; side.
                lit = CSPICE_VSEP(point, subsol_from_inside) LE !DPI/2
;stop,'is it lit?'
                IF NOT lit THEN PLOTS, u_pnt, v_pnt, COLOR = red, PSYM=3
            ENDIF
        ENDFOR
    ENDFOR
;stop,'sahdow'
;; Now do shadow of Saturn on rings if target is Saturn.
    nshadpoints = 40
    IF target EQ 'SATURN' THEN BEGIN
        ;; For these purposes, use approximate inner edge of D ring
        ;; and outer edge of A ring to bound the shadow region radially:
        rin = 67000.
        rout = 136700.
        urange = uringmax-uringmin
        vrange = vringmax-vringmin
        delta_u = urange/nshadpoints
        delta_v = vrange/nshadpoints
        
        CSPICE_NVP2PL, [0D0, 0D0, 1D0], [0D0, 0D0, 0D0], ringplane
        CSPICE_MTXV, TRANSPOSE(inert2p), scpos, scpos_body
        FOR u_ix = 0, nshadpoints-1 DO BEGIN
            u_pnt = uringmin+u_ix*delta_u
            FOR v_ix = 0, nshadpoints-1 DO BEGIN
                v_pnt = vringmin+v_ix*delta_v
                ;; Convert this UV point to RA/Dec. Then get vector from Saturn
                ;; center to that ray where it crosses the ring plane. Then
                ;; see if the ring plane crossing distance is within [rin,rout].
                ;; Then get a vector from that point to the Sun and see if 
                ;; it intersects Saturn.
                UV_2_RD, u_pnt, v_pnt, ra_pnt, dec_pnt, cam_mat
                CSPICE_RADREC, 1D0, ra_pnt*!DTOR, dec_pnt*!DTOR, rec_pnt
                CSPICE_MTXV, TRANSPOSE(inert2p), rec_pnt, direction_body
                CSPICE_INRYPL, -scpos, rec_pnt, ringplane, $
                               nxpts, ringpoint
;            CSPICE_MTXV, inert2p, ringpoint, ringpoint_inert
                ;; ringpoint_inert is vector from Saturn center to ring plane
                ;; intercept point for this particular UV value, in the J2000
                ;; frame. 
                ringdist = CSPICE_VNORM(ringpoint)
                IF (ringdist GE rin) AND (ringdist LE rout) THEN BEGIN
                    ;; Get position of Sun:
                    CSPICE_SPKPOS, 'SATURN', time, 'J2000', 'NONE', 'SUN', $
                                   sunpos, ltimesun
                    view = sunpos - ringpoint
                    ;; view is vector from Sun toward ring plane point.
                    dist2point = CSPICE_VNORM(view)
                    CSPICE_SRFXPT, 'ELLIPSOID', 'SATURN', time, 'NONE', $
                                   'SUN', 'J2000', view, sunpoint, blockdist, $
                                   trgepc, obspos, sunfound
                    CSPICE_SRFXPT, 'ELLIPSOID', 'SATURN', time, 'NONE', $
                                   'CASSINI', 'J2000', rec_pnt, caspoint, $
                                   casblockdist, trgepc, obspos, casfound
                    IF sunfound AND (dist2point LT blockdist) AND $
                      (NOT casfound) THEN $
                        PLOTS, u_pnt, v_pnt, COLOR = red, PSYM =3
                ENDIF
            ENDFOR
        ENDFOR
    ENDIF
ENDIF
;; *****************************OTHER BODY LIMBS***************************
;; Draw the limbs of all bodies:
body_codes = LONG([10, 399, 599, 699, 601, 602, 603, 604, 605, $
    606, 607, 608, 609, 610, 611, 616, 617, 632, 633, 666])
nbodies = N_ELEMENTS(body_codes)

for bodycount = 0, nbodies-1 DO BEGIN
    CSPICE_BODC2N, body_codes[bodycount], bodyname, found
    IF found AND bodyname NE target THEN BEGIN
        BODY_LIMB, bodyname, 'CASSINI', time, 36, ra_limb, $
                   dec_limb, limberror
        IF NOT limberror THEN BEGIN
            RD_2_UV, [ra_limb, ra_limb[0]], [dec_limb, dec_limb[0]], $
                     u_limb, v_limb, behind, cam_mat
            IF NOT behind THEN BEGIN
;; Need to put a test in here to see if the body is behind the main
;; target body.
              OPLOT, u_limb, v_limb
              XYOUTS, MEAN(u_limb), MEAN(v_limb), bodyname, SIZE = labelfontsize
          ENDIF
      ENDIF
  ENDIF
ENDFOR

;;*************************************************************************
;;                                    GRID
;;*************************************************************************
IF KEYWORD_SET(GRID) THEN BEGIN
;; Now draw some RA/Dec gridlines:
;; The RA/Dec grid is distorted in the perspective view of Target, so
;; we need vectors for each line of constant RA and constant Dec.
grid_pnts = 40
racoords = FLTARR(grid_pnts)

ngrid = 5
;; Shift to nearest interval of 0.1 degrees to avoid label clutter:
ramingrid = ROUND(ra_min*10.)/10.
ramaxgrid = ROUND(ra_max*10.)/10.
ra_inc = ROUND((ramaxgrid-ramingrid)/FLOAT(ngrid)*10.)/10.

decmingrid = ROUND(dec_min*10.)/10.
decmaxgrid = ROUND(dec_max*10.)/10.
dec_inc = ROUND((decmaxgrid-decmingrid)/FLOAT(ngrid)*10.)/10.

;; Make sure we have enough grid lines to get to the edge of the plot,
;; after doing the rounding above.
IF ramingrid+ra_inc*(ngrid) LE ramaxgrid THEN nra = ngrid+1 $
  ELSE nra = ngrid
IF decmingrid+dec_inc*(ngrid) LE decmaxgrid THEN ndec = ngrid+1 $
  ELSE ndec = ngrid

;; Calculate the U,V coordinates of the grid lines.
FOR ragrid = 0, nra-1 DO BEGIN
    racoords[*] = ramingrid+ra_inc*ragrid
    deccoords = FINDGEN(grid_pnts)/(grid_pnts-1.)*(decmaxgrid-decmingrid)+$
                decmingrid
    RD_2_UV, racoords, deccoords, ucoords, vcoords, behind, cam_mat
    OPLOT, ucoords, vcoords, COLOR = green
    ;;We've rounded the values to nearest 0.1 degree, so just 
    ;;write the label to that precision as well.
    gridstring = STRTRIM(STRING(racoords[0]),2)
    decimalloc = STRPOS(gridstring,'.')
    gridstring = STRMID(gridstring,0,decimalloc+2)
    XYOUTS, ucoords[grid_pnts-1], vcoords[grid_pnts-1], gridstring, SIZE = labelfontsize
ENDFOR

FOR decgrid = 0, ndec-1 DO BEGIN
    racoords = FINDGEN(grid_pnts)/(grid_pnts-1.)*(ramaxgrid-ramingrid)+$
               ramingrid
    deccoords[*] = decmingrid+dec_inc*decgrid
    RD_2_UV, racoords, deccoords, ucoords, vcoords, behind, cam_mat
    OPLOT, ucoords, vcoords, COLOR = green
    ;;We've rounded the values to nearest 0.1 degree, so just 
    ;;write the label to that precision as well.
    gridstring = STRTRIM(STRING(deccoords[0]),2)
    decimalloc = STRPOS(gridstring,'.')
    gridstring = STRMID(gridstring,0,decimalloc+2)
    XYOUTS, ucoords[grid_pnts-1]+0.01, vcoords[grid_pnts-1]+0.004, gridstring, SIZE = labelfontsize
ENDFOR
ENDIF  ;; Plot the grid or not.

;;**************************************************************************
;;                                      STARS
;;**************************************************************************
IF stars GT 0. THEN BEGIN
starlim = stars
;; Load the star data file:
starfile = FILE_WHICH('star_table.dat')
IF FILE_TEST(starfile) THEN BEGIN
    RESTORE, starfile
    sp_fields = STRMID(STRTRIM(star_table.sp,2),0,1)
    ostarids = WHERE(sp_fields EQ 'O')
    ostars = star_table[ostarids]
    bstarids = WHERE(sp_fields EQ 'B')
    bstars = star_table[bstarids]
    astarids = WHERE(sp_fields EQ 'A')
    astars = star_table[astarids]
    otherstars = star_table[WHERE((sp_fields NE 'A') AND (sp_fields NE 'B') $
                                  AND (sp_fields NE 'O'))]
    
    bright_o = WHERE(ostars.mag LE 1.0 AND ostars.mag LT starlim)
    bright2_o = WHERE(ostars.mag GT 1.0 AND ostars.mag LE 2.0 AND $
        ostars.mag LT starlim)
    mid_o = WHERE(ostars.mag GT 2.0 AND ostars.mag LE 3.0 AND $
        ostars.mag LT starlim)
    faint_o = WHERE(ostars.mag GT 3.0 AND ostars.mag LE starlim)
    
    bright_b = where(bstars.mag LE 1.0 AND bstars.mag LT starlim)
    bright2_b = WHERE(bstars.mag GT 1.0 AND bstars.mag LE 2.0 AND $
        bstars.mag LT starlim)
    mid_b = WHERE(bstars.mag GT 2.0 AND bstars.mag LE 3.0 AND $
        bstars.mag LT starlim)
    faint_b = WHERE(bstars.mag GT 3.0 AND bstars.mag LE starlim AND $
        bstars.mag LT starlim)
        
    bright_a = WHERE(astars.mag LE 1.0 AND astars.mag LT starlim)
    bright2_a = WHERE(astars.mag GT 1.0 AND astars.mag LE 2.0 AND $
        astars.mag LT starlim)
    mid_a = WHERE(astars.mag GT 2.0 AND astars.mag LE 3.0 AND $
        astars.mag LT starlim)
    faint_a = WHERE(astars.mag GT 3.0 AND astars.mag LE starlim)
    
    ;;O Stars*************************************************************
    IF bright_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright_o)-1 DO BEGIN
            RD_2_UV, (ostars[bright_o].ra)[star], $
                     (ostars[bright_o].dec)[star], $
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar,u[0]]
                vstar = [vstar,v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, SYMSIZE = 0.75, $
                   COLOR = blue
    ENDIF

    IF bright2_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright2_o)-1 DO BEGIN
            RD_2_UV, (ostars[bright2_o].ra)[star], $
                     (ostars[bright2_o].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = blue, $
                   SYM_SIZE = .75
    ENDIF

    IF mid_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(mid_o)-1 DO BEGIN
            RD_2_UV, (ostars[mid_o].ra)[star], (ostars[mid_o].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = blue, $
                   SYMSIZE = .75
    ENDIF

    IF faint_o[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(faint_o)-1 DO BEGIN
            RD_2_UV, (ostars[faint_o].ra)[star], (ostars[faint_o].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = blue,$
                   SYMSIZE = .75
    ENDIF

    ;;B Stars*************************************************************
    IF bright_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright_b)-1 DO BEGIN
            RD_2_UV, (bstars[bright_b].ra)[star], $
                     (bstars[bright_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = green, $
                   SYMSIZE = .75
    ENDIF

    IF bright2_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright2_b)-1 DO BEGIN
            RD_2_UV, (bstars[bright2_b].ra)[star], $
                     (bstars[bright2_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = green, $
                   SYMSIZE = .75
    ENDIF

    IF mid_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(mid_b)-1 DO BEGIN
            RD_2_UV, (bstars[mid_b].ra)[star], (bstars[mid_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = yellow, $
                   SYMSIZE = .3
    ENDIF

    IF faint_b[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(faint_b)-1 DO BEGIN
            RD_2_UV, (bstars[faint_b].ra)[star], (bstars[faint_b].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = yellow, $
                   SYMSIZE = .2
    ENDIF

    ;;A Stars*************************************************************
    IF bright_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright_a)-1 DO BEGIN
            RD_2_UV, (astars[bright_a].ra)[star], $
                     (astars[bright_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 1, COLOR = orange, $
                   SYMSIZE = .7
    ENDIF

    IF bright2_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(bright2_a)-1 DO BEGIN
            RD_2_UV, (astars[bright2_a].ra)[star], $
                     (astars[bright2_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = orange, $
                   SYMSIZE = .5
    ENDIF

    IF mid_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(mid_a)-1 DO BEGIN
            RD_2_UV, (astars[mid_a].ra)[star], (astars[mid_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = red, $
                   SYMSIZE = .3
    ENDIF

    IF faint_a[0] NE -1 THEN BEGIN
        ustar = [0]
        vstar = [0]
        FOR star = 0, N_ELEMENTS(faint_a)-1 DO BEGIN
            RD_2_UV, (astars[faint_a].ra)[star], (astars[faint_a].dec)[star],$
                     u, v, behind, cam_mat
            IF NOT behind THEN BEGIN
                ustar = [ustar, u[0]]
                vstar = [vstar, v[0]]
            ENDIF
        ENDFOR
        IF N_ELEMENTS(ustar) GT 1 THEN $
            OPLOT, ustar[1:*], vstar[1:*], PSYM = 8, COLOR = red, $
                   SYMSIZE = .2
    ENDIF

    ;;All other spectral types:
    OPLOT, otherstars.ra, otherstars.dec, PSYM=8, COLOR = red, SYMSIZE=.2

ENDIF ;; If star file found
ENDIF ;; If star keyword set

;;**************************************************************************
;; Plot pixels.
;;**************************************************************************
IF cube[0].et GT 1E19 THEN BEGIN
    ;; cube.et is set to 1E20 in GEOMETER if cube data is not to be used.
    ;; Using data from the GEOMETER data structure in the current session.
    IF KEYWORD_SET(ALLTIMES) THEN BEGIN
        ntimes = N_ELEMENTS(plot_ets)
        FOR time_ix = 0, ntimes-1 DO BEGIN
            ;; Get the camera rotation matrix for this FOV time:
            CSPICE_SPKPOS, target, plot_ets[time_ix], 'J2000', 'LT', $
                           'CASSINI', targpos, ltime
            CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, ratarg, dectarg
            CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, $
                          3L, 2L, 3L, cam_mat_px

            RD_2_UV, [data[time_ix].ra_full[1:*,0],$
                      data[time_ix].ra_full[1,0]], $
                     [data[time_ix].dec_full[1:*,0],$
                      data[time_ix].dec_full[1,0]], u_temp, v_temp, $
                     behind, cam_mat_px
            OPLOT, u_temp, v_temp
    
            IF npixel GT 1 THEN BEGIN
                FOR px_index = 1, npixel-1 DO BEGIN
                    RD_2_UV, [data[time_ix].ra_full[1:*,px_index], $
                            data[time_ix].ra_full[1,px_index]], $
                           [data[time_ix].dec_full[1:*,px_index], $
                            data[time_ix].dec_full[1,px_index]], u_temp, $
                             v_temp, behind, cam_mat_px
                    OPLOT, u_temp, v_temp
                ENDFOR
            ENDIF
            IF NAC THEN BEGIN
              RD_2_UV, [data[time_ix].ra_full_nac[1:*], data[time_ix].ra_full_nac[1]], $
                [data[time_ix].dec_full_nac[1:*], data[time_ix].dec_full_nac[1]], $
                u_tempnac, v_tempnac, behind, cam_mat_px
              OPLOT, u_tempnac, v_tempnac
            ENDIF
            IF WAC THEN BEGIN
              RD_2_UV, [data[time_ix].ra_full_wac[1:*], data[time_ix].ra_full_wac[1]], $
                [data[time_ix].dec_full_wac[1:*], data[time_ix].dec_full_wac[1]], $
                u_tempwac, v_tempwac, behind, cam_mat_px
              OPLOT, u_tempwac, v_tempwac
            ENDIF
        ENDFOR  ;; Looping over times.
;; Just label the last time:
        XYOUTS, u_temp[0], v_temp[0], channel_name, SIZE = labelfontsize
        
    ENDIF ELSE BEGIN
        RD_2_UV, [data[i_time].ra_full[1:*,0],data[i_time].ra_full[1,0]], $
               [data[i_time].dec_full[1:*,0],data[i_time].dec_full[1,0]], $
                 u_temp, v_temp, behind, cam_mat
        OPLOT, u_temp, v_temp
        XYOUTS, u_temp[0], v_temp[0], channel_name, SIZE = labelfontsize
        
        IF NAC THEN BEGIN
          RD_2_UV, [data[i_time].ra_full_nac[1:*], data[i_time].ra_full_nac[1]], $
            [data[i_time].dec_full_nac[1:*], data[i_time].dec_full_nac[1]], $
            u_tempnac, v_tempnac, behind, cam_mat
          OPLOT, u_tempnac, v_tempnac
        ENDIF
        IF WAC THEN BEGIN
          RD_2_UV, [data[i_time].ra_full_wac[1:*], data[i_time].ra_full_wac[1]], $
            [data[i_time].dec_full_wac[1:*], data[i_time].dec_full_wac[1]], $
            u_tempwac, v_tempwac, behind, cam_mat
          OPLOT, u_tempwac, v_tempwac
        ENDIF
        
        IF npixel GT 1 THEN BEGIN
            FOR px_index = 1, npixel-1 DO BEGIN
                RD_2_UV, [data[i_time].ra_full[1:*,px_index], $
                        data[i_time].ra_full[1,px_index]], $
                       [data[i_time].dec_full[1:*,px_index], $
                        data[i_time].dec_full[1,px_index]], u_temp, v_temp, $
                         behind, cam_mat
                OPLOT, u_temp, v_temp

                IF px_index/2.-FIX(px_index/2.) EQ 0. THEN BEGIN
                   ;;Only label every other row.
                   name = 'Row '+STRTRIM(STRING(selected_pixels[px_index]),2)
                   XYOUTS, u_temp[0], v_temp[0], name, SIZE = labelfontsize
                ENDIF
            ENDFOR
;            name = 'Row '+STRTRIM(STRING(selected_pixels[0]),2)
;            XYOUTS, u_temp[0], v_temp[0], name, SIZE = labelfontsize
        ENDIF
    ENDELSE  ;;Only plotting at one selected time.

ENDIF ELSE BEGIN

;; Plotting data from loaded image cube (also computed by GEOMETER, but
;; run and saved at a different time).

    nrows = N_ELEMENTS(WHERE(cube[0].data[*,0] NE -1.))
    ntimes = N_ELEMENTS(plot_ets)
;stop,'1'
;; Figure out colors. Scale each channel to 255.
    IF colormax[0] EQ 0. THEN BEGIN
        redmax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            redmax = redmax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[0]:spec_bounds[1]]))
    ENDIF ELSE redmax = colormax[0]
    IF colormax[1] EQ 0. THEN BEGIN
        greenmax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            greenmax = greenmax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[2]:spec_bounds[3]]))
    ENDIF ELSE greenmax = colormax[1]
    IF colormax[2] EQ 0. THEN BEGIN
        bluemax = 1.E-20
        FOR time_ix = 0, ntimes-1 DO $
          FOR px_index = 0, npixel-1 DO $
            bluemax = bluemax > $
          MAX(TOTAL(cube[time_ix].data[px_index,$
                              spec_bounds[4]:spec_bounds[5]]))
    ENDIF ELSE bluemax = colormax[2]

    pixelstodraw = WHERE(maininfo.pxvec)
    ;; If no individual pixels have been selected by the user, assume
    ;; the user wants all pixels drawn:
    IF pixelstodraw[0] EQ -1 THEN pixelstodraw = INDGEN(64)
    npixel = N_ELEMENTS(pixelstodraw)

    IF KEYWORD_SET(ALLTIMES) THEN BEGIN
        FOR time_ix = 0, ntimes-1 DO BEGIN
            ;; Get the camera rotation matrix for this FOV time:
            CSPICE_SPKPOS, target, plot_ets[time_ix], 'J2000', 'LT', $
                           'CASSINI', targpos, ltime
            CSPICE_RECRAD, targpos/CSPICE_VNORM(targpos), unit, ratarg, dectarg
            CSPICE_EUL2M, !DPI/2D0, !DPI/2D0-dectarg, ratarg, $
                          3L, 2L, 3L, cam_mat_px

            FOR px_number = 0, npixel-1 DO BEGIN
               px_index = pixelstodraw[px_number]
               ;; Build vectors for plotting the boundary of the slit:
                ravec = [cube[time_ix].ra_full[*,px_index], $
                         cube[time_ix].ra_full[0,px_index]]
                decvec = $
                  [cube[time_ix].dec_full[*,px_index], $
                   cube[time_ix].dec_full[0,px_index]]

                RD_2_UV, ravec, decvec, u_temp, v_temp, behind, cam_mat_px
                ;;Get the brightness of the pixel:
                redcolor = TOTAL(cube[time_ix].data[px_index,$
                                           spec_bounds[0]:spec_bounds[1]]) $
                           < redmax
                redcolor /= redmax
                redcolor *= 255.

                greencolor = TOTAL(cube[time_ix].data[px_index,$
                                             spec_bounds[2]:spec_bounds[3]]) $
                             < greenmax
                greencolor /= greenmax
                greencolor *= 255.

                bluecolor = TOTAL(cube[time_ix].data[px_index,$
                                            spec_bounds[4]:spec_bounds[5]]) $
                            < bluemax
                bluecolor /= bluemax
                bluecolor *= 255.

                ;; Now combine these three shades into a single color:
;; problem: PS doesn't accept TrueColor plots. Only images. 
                TVLCT, INDGEN(256), INDGEN(256), INDGEN(256)

                pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                          256L*256L*LONG(bluecolor)
                OPLOT, u_temp, v_temp, COLOR = pxcolor
                POLYFILL, u_temp, v_temp, COLOR = pxcolor
            ENDFOR  ;; Building pixel vectors.
        ENDFOR  ;; Looping over times.
;; Just label the last time:
        XYOUTS, u_temp[0], v_temp[0], channel_name, SIZE = labelfontsize
        
    ENDIF ELSE BEGIN
;; Just plotting at one selected time.
        FOR px_number = 0, npixel-1 DO BEGIN
           px_index = pixelstodraw[px_number]
            ravec = [cube[i_time].ra_full[*,px_index], $
                     cube[i_time].ra_full[0,px_index]]
            
            decvec = $
              [cube[i_time].dec_full[*,px_index], $
               cube[i_time].dec_full[0,px_index]]

            RD_2_UV, ravec, decvec, u_temp, v_temp, behind, cam_mat

            ;;Get the color of the pixel:
            redcolor = TOTAL(cube[i_time].data[px_index,$
                                       spec_bounds[0]:spec_bounds[1]]) $
                       < redmax
            redcolor /= redmax
            redcolor *= 255.
            
            greencolor = TOTAL(cube[i_time].data[px_index,$
                                         spec_bounds[2]:spec_bounds[3]]) $
                         < greenmax
            greencolor /= greenmax
            greencolor *= 255.
            
            bluecolor = TOTAL(cube[i_time].data[px_index,$
                                        spec_bounds[4]:spec_bounds[5]]) $
                        < bluemax
            bluecolor /= bluemax
            bluecolor *= 255.
            
            ;; Now combine these three shades into a single color:
            pxcolor = LONG(redcolor)+256L*LONG(greencolor)+$
                      256L*256L*LONG(bluecolor)

            OPLOT, u_temp, v_temp, COLOR = pxcolor
            POLYFILL, u_temp, v_temp, COLOR = pxcolor
            IF px_index/2 EQ px_index THEN BEGIN
               ;; Only label every other row
               name = 'Row '+STRTRIM(STRING(px_index),2)
               XYOUTS, u_temp[0], v_temp[0], name, SIZE = labelfontsize
            ENDIF
        ENDFOR
;        name = 'Row '+STRTRIM(STRING(px_index-1),2)
;        XYOUTS, u_temp[0], v_temp[0], name, SIZE = labelfontsize
    ENDELSE  ;;Only plotting at one selected time.
    
ENDELSE  ;;Plotting data cube pixels.

IF KEYWORD_SET(TOPGRID) AND KEYWORD_SET(LATLON) THEN BEGIN
;; Redraw the lat/lon grid.

FOR lat_ix = 0, nlatlines-1 DO BEGIN
    latvec = DBLARR(ndrawpoints)+lat_min+lat_ix*lat_int
    lonvec = DINDGEN(ndrawpoints)/(ndrawpoints-1D0)*!DPI*2D0
    CSPICE_LATREC, DBLARR(ndrawpoints)+radii[0], lonvec, latvec, $
                   latlonpos

    FOR vec_ix = 0, ndrawpoints-1 DO BEGIN
        latlonvec = REFORM(latlonpos[*,vec_ix])
        latlonvec /= CSPICE_VNORM(latlonvec)
        latlonvec *= SQRT((radii[0]*COS(latvec[lat_ix]))^2+$
                          (radii[2]*SIN(latvec[lat_ix]))^2)
        ;; latlonvec is the vector to the particular point at this 
        ;; latitude and longitude on the surface of oblate spheroid
        ;; target (note, not coded for triaxial ellipsoid, just oblate
        ;; spheroid).
        CSPICE_MTXV, inert2platlon, latlonvec, latlonvec_inert
        vec2point = latlonvec_inert-scposlatlon
        IF CSPICE_VNORM(vec2point) LE CSPICE_VNORM(scposlatlon) THEN $
          visible_vec[vec_ix] = 1 ELSE visible_vec[vec_ix] = 0
        CSPICE_RECRAD, vec2point, dist, ra_pnt, dec_pnt
        RD_2_UV, ra_pnt*!RADEG, dec_pnt*!RADEG, u_pnt, v_pnt, behind, $
                 cam_mat_latlon
        u_lat_vec[vec_ix] = u_pnt
        v_lat_vec[vec_ix] = v_pnt
    ENDFOR
    good = WHERE(visible_vec)
    IF N_ELEMENTS(good) GT 2 THEN BEGIN
        delta_good = good-SHIFT(good,1)
        junk = MAX(delta_good, breakloc)
        good_us = u_lat_vec[good[0:breakloc-1]]
        good_vs = v_lat_vec[good[0:breakloc-1]]
        latlabel_u = MEAN(good_us)
        junk = MIN(ABS(latlabel_u-good_us), loc)
        latlabel_v = good_vs[loc]
        IF lat_ix EQ FIX(nlatlines/2) THEN $
          OPLOT, good_us, good_vs, COLOR = white ELSE $
            OPLOT, good_us, good_vs, COLOR = cyan
        good_us = u_lat_vec[good[breakloc:*]]
        good_vs = v_lat_vec[good[breakloc:*]]
        IF lat_ix EQ FIX(nlatlines/2) THEN $
          OPLOT, good_us, good_vs, COLOR = white ELSE $
            OPLOT, good_us, good_vs, COLOR = cyan
        IF lat_ix EQ FIX(nlatlines/2-1) THEN $
          XYOUTS, latlabel_u, latlabel_v, $
                  STRMID(STRING(STRTRIM(ROUND(lat_int*!RADEG),2)),0,2), SIZE=labelfontsize ELSE $
                    IF lat_ix EQ FIX(nlatlines/2+1) THEN $
                      XYOUTS, latlabel_u, latlabel_v, '-'+$
          STRMID(STRING(STRTRIM(ROUND(lat_int*!RADEG),2)),0,2), SIZE = labelfontsize
    ENDIF
ENDFOR

;; Same thing for longitude:
FOR lon_ix = 0, nlonlines-1 DO BEGIN
    lonvec = DBLARR(ndrawpoints)+lon_min+lon_ix*long_int
    latvec = DINDGEN(ndrawpoints)/(ndrawpoints-1D0)*!DPI-!DPI/2D0
    CSPICE_LATREC, DBLARR(ndrawpoints)+radii[0], lonvec, latvec, $
                   latlonpos

    ;; I think I can get rid of the following loop by using vectorized
    ;; CSPICE calls, but just get it working for now and worry about
    ;; performance later.
    FOR vec_ix = 0, ndrawpoints-1 DO BEGIN
        latlonvec = REFORM(latlonpos[*,vec_ix])
        latlonvec /= CSPICE_VNORM(latlonvec)
        latlonvec *= SQRT((radii[0]*COS(latvec[lat_ix]))^2+$
                          (radii[2]*SIN(latvec[lat_ix]))^2)
        ;; latlonvec is the vector to the particular point at this 
        ;; latitude and longitude on the surface of oblate spheroid
        ;; target (note, not coded for triaxial ellipsoid, just oblate
        ;; spheroid).
        CSPICE_MTXV, inert2platlon, latlonvec, latlonvec_inert
        vec2point = latlonvec_inert-scposlatlon
        IF CSPICE_VNORM(vec2point) LE CSPICE_VNORM(scposlatlon) THEN $
          visible_vec[vec_ix] = 1 ELSE visible_vec[vec_ix] = 0
        CSPICE_RECRAD, vec2point, dist, ra_pnt, dec_pnt
        RD_2_UV, ra_pnt*!RADEG, dec_pnt*!RADEG, u_pnt, v_pnt, behind, $
                 cam_mat_latlon
        u_lon_vec[vec_ix] = u_pnt
        v_lon_vec[vec_ix] = v_pnt
    ENDFOR
    good = WHERE(visible_vec)
    IF N_ELEMENTS(good) GT 2 THEN BEGIN
        delta_good = good-SHIFT(good,1)
        junk = MAX(delta_good, breakloc)
        good_us = u_lon_vec[good[0:breakloc-1]]
        good_vs = v_lon_vec[good[0:breakloc-1]]
        OPLOT, good_us, good_vs, COLOR = cyan
        good_us = u_lon_vec[good[breakloc:*]]
        good_vs = v_lon_vec[good[breakloc:*]]
        OPLOT, good_us, good_vs, COLOR = cyan
        lon_v_label = MEAN(good_vs)
        junk = MIN(ABS(lon_v_label-good_vs), loc)
;        IF lon_ix EQ 0 THEN $
;          XYOUTS, good_us[loc], lon_v_label, '0' , SIZE = labelfontsize ELSE $
;            IF lon_ix EQ FIX(nlonlines/2-1) THEN $
;              XYOUTS, good_us[loc], lon_v_label, '180', SIZE = labelfontsize
    ENDIF
ENDFOR

ENDIF

;; Do some labels:
   CSPICE_ET2UTC, time, 'ISOD', 3L, time_label
;XYOUTS, u_min-ABS(u_min-u_max)/50., v_min-(v_max-v_min)/20., $
;        time_label, SIZE = labelfontsize
;XYOUTS, u_min-ABS(u_min-u_max)/2., v_min-(v_max-v_min)/20., $
;        STRTRIM(STRING(CSPICE_VNORM(targpos2)),2)+' km',SIZE=labelfontsize

   ra_out = STRTRIM(STRING(ratarg*!RADEG, '(F8.2)'), 2)
   dec_out = STRTRIM(STRING(dectarg*!RADEG, '(F8.2)'), 2)

   CSPICE_SUBPT, 'nearpoint', target, time, 'LT', $
                 'CASSINI', point, alt
   CSPICE_RECLAT, point, radius, scposlon, scposlat
   scposlon *= !RADEG
   scposlat *= !RADEG

   CSPICE_SUBSOL, 'nearpoint', target, time, 'LT', $
                  'CASSINI', sunpoint
   CSPICE_RECLAT, sunpoint, radius, subsollon, subsollat
   subsollon *= !RADEG
   subsollat *= !RADEG

   scpos_out = STRTRIM(STRING(scposlat, '(D8.2)'), 2) + ', ' + $
               STRTRIM(STRING(scposlon, '(D8.2)'), 2)

   subsolar_out = STRTRIM(STRING(subsollat, '(D8.2)'), 2) + ', ' + $
               STRTRIM(STRING(subsollon, '(D8.2)'), 2)

IF (KEYWORD_SET(SCREEN) OR KEYWORD_SET(JPG) OR KEYWORD_SET(TIFF)) THEN BEGIN
   XYOUTS, 10, 70, time_label + '  ' + $
           STRTRIM(STRING(CSPICE_VNORM(targpos2)),2)+' km', $
           SIZE = labelfontsize, /DEVICE
   XYOUTS, 10, 50, 'Target RA/dec: '+ ra_out + ', ' + dec_out, $
           SIZE = labelfontsize, /DEVICE
   XYOUTS, 10, 30, 'Subsolar lat/lon: ' + subsolar_out, SIZE = labelfontsize, /DEVICE
   XYOUTS, 10, 10, 'Sub-s/c lat/lon: ' + scpos_out, SIZE = labelfontsize, /DEVICE
ENDIF
IF KEYWORD_SET(POSTSCRIPT) THEN BEGIN
   XYOUTS, 50, 2200, time_label + '  ' + $
           STRTRIM(STRING(CSPICE_VNORM(targpos2)),2)+' km', $
           SIZE = labelfontsize, /DEVICE
   XYOUTS, 50, 1500, 'Target RA/dec: '+ ra_out + ', ' + dec_out, $
           SIZE = labelfontsize, /DEVICE
   XYOUTS, 50, 800, 'Subsolar lat/lon: ' + subsolar_out, SIZE = labelfontsize, /DEVICE
   XYOUTS, 50, 100, 'Sub-s/c lat/lon: ' + scpos_out, SIZE = labelf, /DEVICE
ENDIF

;; Write JPEG or TIFF files:

IF KEYWORD_SET(JPG) THEN BEGIN
   DEVICE, GET_DECOMPOSE = old_decomposed
   DEVICE, DECOMPOSED = 0
   DEVICE, RETAIN=2
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.jpg') AND (extension NE 'jpeg')) THEN $
      savefile = savefile + '.jpg'
   WRITE_JPEG, savefile, win, QUALITY = 100, TRUE = 1
   DEVICE, DECOMPOSED = old_decomposed
ENDIF

IF KEYWORD_SET(TIFF) THEN BEGIN
  DEVICE, GET_DECOMPOSE = old_decomposed
  DEVICE, DECOMPOSED = 0
  DEVICE, RETAIN=2
   win = TVRD(TRUE = 1)
   extension = STRMID(savefile, 3, /REVERSE_OFFSET)
   IF ((extension NE '.tif') AND (extension NE 'tiff')) THEN $
      savefile = savefile + '.tif'
   WRITE_TIFF, savefile, win, ORIENTATION = 0
   DEVICE, DECOMPOSED = old_decomposed
ENDIF


IF KEYWORD_SET(POSTSCRIPT) THEN BEGIN
   DEVICE, /CLOSE
ENDIF


END
